# -*- coding: utf8 -*-
"""
Burp-UI is a web-ui for burp backup written in python with Flask and
jQuery/Bootstrap

.. module:: burpui.cli
    :platform: Unix
    :synopsis: Burp-UI CLI module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>
"""
import os
import sys
import click

if os.getenv('BUI_MODE') in ['server', 'ws'] or 'websocket' in sys.argv:
    try:
        from gevent import monkey
        monkey.patch_socket()
    except ImportError:
        pass

from .app import create_app  # noqa
from .exceptions import BUIserverException  # noqa
from six import iteritems  # noqa

try:
    from flask_socketio import SocketIO  # noqa
    WS_AVAILABLE = True
except ImportError:
    WS_AVAILABLE = False

ROOT = os.path.dirname(os.path.realpath(__file__))
DEBUG = os.getenv('BUI_DEBUG') or os.getenv('FLASK_DEBUG') or False
DEBUG = DEBUG and DEBUG.lower() not in ['false', 'no', '0']

VERBOSE = os.getenv('BUI_VERBOSE') or 0
if VERBOSE:
    try:
        VERBOSE = int(VERBOSE)
    except ValueError:
        VERBOSE = 0

# UNITTEST is used to skip the burp-2 requirements for modes != server
UNITTEST = os.getenv('BUI_MODE') not in ['server', 'manage', 'celery', 'legacy', 'ws']
CLI = os.getenv('BUI_MODE') not in ['server', 'legacy', 'ws']

try:
    app = create_app(
        conf=os.environ.get('BUI_CONFIG'),
        verbose=VERBOSE,
        logfile=os.environ.get('BUI_LOGFILE'),
        debug=DEBUG,
        gunicorn=False,
        unittest=UNITTEST,
        cli=CLI,
        websocket_server=(os.getenv('BUI_MODE') == 'ws' or 'websocket' in sys.argv)
    )
except:
    import traceback
    traceback.print_exc()
    sys.exit(1)

try:
    from .app import create_db
    from .ext.sql import db
    from flask_migrate import Migrate

    # This may have been reseted by create_app
    if isinstance(app.database, bool):
        app.config['WITH_SQL'] = app.database
    else:
        app.config['WITH_SQL'] = app.database and \
            app.database.lower() != 'none'

    if app.config['WITH_SQL']:
        create_db(app, True)

        mig_dir = os.getenv('BUI_MIGRATIONS')
        if mig_dir:
            migrate = Migrate(app, db, mig_dir)
        else:
            migrate = Migrate(app, db)
except ImportError:
    pass


def _die(error, appli=None):
    appli = " '{}'".format(appli) if appli else ''
    click.echo(
        click.style(
            'Unable to initialize the application{}: {}'.format(appli, error),
            fg='red'
        ),
        err=True
    )
    sys.exit(2)


@app.cli.command()
def legacy():
    """Legacy server for backward compatibility."""
    click.echo(
        click.style(
            'If you want to pass options, you should run \'python -m burpui '
            '-m legacy [...]\' instead',
            fg='yellow'
        )
    )
    app.manual_run()


@app.cli.command()
@click.option('-b', '--bind', default='127.0.0.1',
              help='Which address to bind to for the websocket server')
@click.option('-p', '--port', default=5001,
              help='Which port to listen on for the websocket server')
@click.option('-d', '--debug', default=False, is_flag=True,
              help='Whether to start the websocket server in debug mode')
def websocket(bind, port, debug):
    """Start a new websocket server."""
    try:
        from .ext.ws import socketio
    except ImportError:
        _die('Missing requirement, did you ran \'pip install'
             ' "burp-ui[websocket]"\'?', 'websocket')
    socketio.run(app, host=bind, port=port, debug=debug)


@app.cli.command()
@click.option('-b', '--backend', default='BASIC',
              help='User Backend (default is BASIC).')
@click.option('-p', '--password', help='Password to assign to user.',
              default=None)
@click.option('-a', '--ask', default=False, is_flag=True,
              help='If no password is provided and this flag is enabled, '
                   'you\'ll be prompted for one, else a random one will be '
                   'generated.')
@click.option('-v', '--verbose', default=False, is_flag=True,
              help='Add extra debug messages.')
@click.argument('name')
def create_user(backend, password, ask, verbose, name):
    """Create a new user."""
    try:
        msg = app.load_modules(True)
    except Exception as e:
        msg = str(e)

    if msg:
        _die(msg, 'create_user')

    click.echo(click.style('[*] Adding \'{}\' user...'.format(name), fg='blue'))
    try:
        handler = getattr(app, 'uhandler')
    except AttributeError:
        handler = None

    if not handler or len(handler.backends) == 0 or \
            backend not in handler.backends:
        click.echo(click.style('[!] No authentication backend found', fg='red'))
        sys.exit(1)

    back = handler.backends[backend]

    if back.add_user is False:
        click.echo(click.style("[!] The '{}' backend does not support user "
                               "creation".format(backend), fg='red'))
        sys.exit(2)

    if not password:
        if ask:
            import getpass
            password = getpass.getpass()
            confirm = getpass.getpass('Confirm: ')
            if password != confirm:
                click.echo(click.style("[!] Passwords mismatch", fg='red'))
                sys.exit(3)
        else:
            import random

            alphabet = "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLM" \
                       "NOPQRSTUVWXYZ"
            pw_length = 8
            mypw = ""

            for i in range(pw_length):
                next_index = random.randrange(len(alphabet))
                mypw += alphabet[next_index]
            password = mypw
            click.echo(
                click.style(
                    '[+] Generated password: {}'.format(password),
                    fg='blue'
                )
            )

    success, message, _ = back.add_user(name, password)
    click.echo(click.style(
        '[+] Success: {}{}'.format(
            success, ' -> {}'.format(message) if verbose and message else ''
        ),
        fg='green' if success else 'red')
    )


@app.cli.command()
@click.option('-p', '--password', help='Password to assign to user.',
              default=None)
@click.option('-u', '--username', help='Provide the username to get the full '
              'configuration line.',
              default=None)
@click.option('-b', '--batch', default=False, is_flag=True,
              help='Don\'t be extra verbose so that you can use the output '
              'directly in your scripts. Requires both -u and -p.')
def hash_password(password, username, batch):
    """Hash a given password to fill the configuration file."""
    from werkzeug.security import generate_password_hash

    if batch and (not username or not password):
        click.echo(click.style(
            'You need to provide both a username and a password using the '
            '-u and -p flags!',
            fg='red')
        )
        sys.exit(1)

    askpass = False
    if not password:
        askpass = True
        import getpass
        password = getpass.getpass()

    hashed = generate_password_hash(password)
    if not batch:
        click.echo("'{}' hashed into: {}".format(
            password if not askpass else '*' * 8,
            hashed)
        )
    if username:
        if not batch:
            click.echo(click.style(
                '#8<{}'.format('-' * 77),
                fg='blue')
            )
        click.echo('{} = {}'.format(username, hashed))
        if not batch:
            click.echo(click.style(
                '#8<{}'.format('-' * 77),
                fg='blue')
            )


@app.cli.command()
@click.argument('language')
def init_translation(language):
    """Initialize a new translation for the given language."""
    try:
        import babel  # noqa
    except ImportError:
        click.echo(
            click.style('Missing i18n requirements, giving up', fg='yellow')
        )
        return
    os.chdir(os.path.join(ROOT, '..'))
    os.system('pybabel extract -F babel.cfg -k __ -k lazy_gettext -o messages.pot burpui')
    os.system('pybabel init -i messages.pot -d burpui/translations -l {}'.format(language))
    os.unlink('messages.pot')


@app.cli.command()
def update_translation():
    """Update translation files."""
    try:
        import babel  # noqa
    except ImportError:
        click.echo(
            click.style('Missing i18n requirements, giving up', fg='yellow')
        )
        return
    os.chdir(os.path.join(ROOT, '..'))
    os.system('pybabel extract -F babel.cfg -k __ -k lazy_gettext -o messages.pot burpui')
    os.system('pybabel update -i messages.pot -d burpui/translations')
    os.unlink('messages.pot')


@app.cli.command()
def compile_translation():
    """Compile translations."""
    try:
        import babel  # noqa
    except ImportError:
        click.echo(
            click.style('Missing i18n requirements, giving up', fg='yellow')
        )
        return
    os.chdir(os.path.join(ROOT, '..'))
    os.system('pybabel compile -f -d burpui/translations')


@app.cli.command()
@click.option('-b', '--burp-conf-cli', 'bconfcli', default=None,
              help='Burp client configuration file')
@click.option('-s', '--burp-conf-serv', 'bconfsrv', default=None,
              help='Burp server configuration file')
@click.option('-c', '--client', default='bui',
              help='Name of the burp client that will be used by Burp-UI '
                   '(defaults to "bui")')
@click.option('-h', '--host', default='::1',
              help='Address of the status server (defaults to "::1")')
@click.option('-r', '--redis', default=None,
              help='Redis URL to connect to')
@click.option('-d', '--database', default=None,
              help='Database to connect to for persistent storage')
@click.option('-p', '--plugins', default=None,
              help='Plugins location')
@click.option('-n', '--dry', is_flag=True,
              help='Dry mode. Do not edit the files but display changes')
def setup_burp(bconfcli, bconfsrv, client, host, redis, database, plugins, dry):
    """Setup burp client for burp-ui."""
    if app.vers != 2:
        click.echo(
            click.style(
                'Sorry, you can only setup the Burp 2 client',
                fg='red'
            ),
            err=True
        )
        sys.exit(1)

    if not app.standalone:
        click.echo(
            click.style(
                'Sorry, only the standalone mode is supported',
                fg='red'
            ),
            err=True
        )
        sys.exit(1)

    try:
        msg = app.load_modules(True)
    except Exception as e:
        msg = str(e)

    if msg:
        _die(msg, 'setup_burp')

    from .misc.parser.utils import Config
    from .app import get_redis_server
    import difflib
    import tempfile

    parser = app.client.get_parser()
    orig = source = None
    conf_orig = []
    if dry:
        try:
            with open(app.conf.options.filename) as fil:
                conf_orig = fil.readlines()
        except:
            pass

        orig = source = app.conf.options.filename
        (_, temp) = tempfile.mkstemp()
        app.conf.options.filename = temp

    # handle migration of old config files
    if app.conf.section_exists('Burp2'):
        if app.conf.rename_section('Burp2', 'Burp', source):
            click.echo(
                click.style(
                    'Renaming old [Burp2] section',
                    fg='blue'
                )
            )
            app.conf._refresh(True)

    refresh = False
    if not app.conf.lookup_section('Burp', source):
        refresh = True
    if not app.conf.lookup_section('Global', source):
        refresh = True
    if (database or redis) and not app.conf.lookup_section('Production', source):
        refresh = True

    if refresh:
        app.conf._refresh(True)

    def _edit_conf(key, val, attr, section='Burp', obj=app.client):
        if val and (((key not in app.conf.options[section]) or
                    (key in app.conf.options[section] and
                    val != app.conf.options[section][key])) and
                    getattr(obj, attr) != val):
            app.conf.options[section][key] = val
            app.conf.options.write()
            click.echo(
                click.style(
                    'Adding new option: "{}={}" to section [{}]'.format(
                        key,
                        val,
                        section
                    ),
                    fg='blue'
                )
            )
            return True
        return False

    def _color_diff(line):
        if line.startswith('+'):
            return click.style(line, fg='green')
        elif line.startswith('-'):
            return click.style(line, fg='red')
        elif line.startswith('^'):
            return click.style(line, fg='blue')
        return line

    refresh = False
    refresh |= _edit_conf('bconfcli', bconfcli, 'burpconfcli')
    refresh |= _edit_conf('bconfsrv', bconfsrv, 'burpconfsrv')
    refresh |= _edit_conf('plugins', plugins, 'plugins', 'Global', app)

    if refresh:
        app.conf._refresh(True)

    if redis:
        try:
            # detect missing modules
            import redis as redis_client  # noqa
            import celery  # noqa
            import socket
            if ('redis' not in app.conf.options['Production'] or
                'redis' in app.conf.options['Production'] and
                app.conf.options['Production']['redis'] != redis) and \
                    app.redis != redis:
                app.conf.options['Production']['redis'] = redis

            rhost, rport, _ = get_redis_server(app)
            ret = -1
            for res in socket.getaddrinfo(rhost, rport, socket.AF_UNSPEC, socket.SOCK_STREAM):
                if ret == 0:
                    break
                af, socktype, proto, _, sa = res
                try:
                    s = socket.socket(af, socktype, proto)
                except socket.error:
                    continue
                try:
                    ret = s.connect_ex(sa)
                except:
                    continue

            if ret == 0:
                app.conf.options['Production']['celery'] = 'true'

                app.conf.options['Production']['storage'] = 'redis'

                app.conf.options['Production']['cache'] = 'redis'
            else:
                click.echo(
                    click.style(
                        'Unable to contact the redis server, disabling it',
                        fg='yellow'
                    )
                )
                app.conf.options['Production']['storage'] = 'default'
                app.conf.options['Production']['cache'] = 'default'
                if app.use_celery:
                    app.conf.options['Production']['celery'] = 'false'

            app.conf.options.write()
            app.conf._refresh(True)
        except ImportError:
            click.echo(
                click.style(
                    'Unable to activate redis & celery. Did you ran the '
                    '\'pip install burp-ui[celery]\' and '
                    '\'pip install burp-ui[gunicorn-extra]\' commands first?',
                    fg='yellow'
                )
            )

    if database:
        try:
            from .ext.sql import db  # noqa
            if ('database' not in app.conf.options['Production'] or
                'database' in app.conf.options['Production'] and
                app.conf.options['Production']['database'] != database) and \
                    app.database != database:
                app.conf.options['Production']['database'] = database
                app.conf.options.write()
                app.conf._refresh(True)
        except ImportError:
            click.echo(
                click.style(
                    'It looks like some dependencies are missing. Did you ran '
                    'the \'pip install "burp-ui[sql]"\' command first?',
                    fg='yellow'
                )
            )

    if dry:
        temp = app.conf.options.filename
        app.conf.options.filename = orig
        after = []
        try:
            if not os.path.exists(temp) or os.path.getsize(temp) == 0:
                after = conf_orig
            else:
                with open(temp) as fil:
                    after = fil.readlines()
                os.unlink(temp)
        except:
            pass
        diff = difflib.unified_diff(conf_orig, after, fromfile=orig, tofile='{}.new'.format(orig))
        out = ''
        for line in diff:
            out += _color_diff(line)
        if out:
            click.echo_via_pager(out)

    bconfcli = bconfcli or app.conf.options['Burp'].get('bconfcli') or \
        getattr(app.client, 'burpconfcli')
    bconfsrv = bconfsrv or app.conf.options['Burp'].get('bconfsrv') or \
        getattr(app.client, 'burpconfsrv')
    dest_bconfcli = bconfcli

    if not os.path.exists(bconfsrv):
        click.echo(
            click.style(
                'Unable to locate burp-server configuration, aborting!',
                fg='red'
            ),
            err=True
        )
        sys.exit(1)

    confsrv = Config(bconfsrv, parser, 'srv')
    confsrv.set_default(bconfsrv)
    confsrv.parse()

    if host not in ['::1', '127.0.0.1']:
        bind = confsrv.get('status_address')
        if (bind and bind not in [host, '::', '0.0.0.0']) or not bind:
            click.echo(
                click.style(
                    'It looks like your burp server is not exposing it\'s '
                    'status port in a way that is reachable by Burp-UI!',
                    fg='yellow'
                )
            )
            click.echo(
                click.style(
                    'You may want to set the \'status_address\' setting with '
                    'either \'{}\', \'::\' or \'0.0.0.0\' in the {} file '
                    'in order to make Burp-UI work'.format(host, bconfsrv),
                    fg='blue'
                )
            )

    status_port = confsrv.get('status_port', [4972])
    if 'max_status_children' not in confsrv:
        click.echo(
            click.style(
                'We need to set the number of \'max_status_children\'. '
                'Setting it to 15.',
                fg='blue'
            )
        )
        confsrv['max_status_children'] = 15
        status_port = status_port[0]
    else:
        max_status_children = confsrv.get('max_status_children')
        found = False
        for idx, value in enumerate(max_status_children):
            if value >= 15:
                found = True
                if idx >= len(status_port):
                    status_port = status_port[-1]
                else:
                    status_port = status_port[idx]
                break
        if not found:
            click.echo(
                click.style(
                    'We need to raise the number of \'max_status_children\'. '
                    'Raising it to 15 instead of {}.'.format(max_status_children),
                    fg='yellow'
                )
            )
            confsrv['max_status_children'][-1] = 15
            status_port = status_port[-1]

    if 'restore_client' not in confsrv:
        confsrv['restore_client'] = client
    else:
        restore = confsrv.getlist('restore_client')
        if client not in restore:
            confsrv['restore_client'].append(client)

    confsrv['monitor_browse_cache'] = True

    ca_client_dir = confsrv.get('ca_csr_dir')
    if ca_client_dir and not os.path.exists(ca_client_dir):
        try:
            os.makedirs(ca_client_dir)
        except IOError as exp:
            click.echo(
                click.style(
                    'Unable to create "{}" dir: {}'.format(ca_client_dir, exp),
                    fg='yellow'
                ),
                err=True
            )

    if confsrv.dirty:
        if dry:
            (_, dstfile) = tempfile.mkstemp()
        else:
            dstfile = bconfsrv

        confsrv.store(conf=bconfsrv, dest=dstfile, insecure=True)
        if dry:
            before = []
            after = []
            try:
                with open(bconfsrv) as fil:
                    before = fil.readlines()
            except:
                pass
            try:
                with open(dstfile) as fil:
                    after = fil.readlines()
                os.unlink(dstfile)
            except:
                pass
            diff = difflib.unified_diff(before, after, fromfile=bconfsrv, tofile='{}.new'.format(bconfsrv))
            out = ''
            for line in diff:
                out += _color_diff(line)
            if out:
                click.echo_via_pager(out)

    if confsrv.get('clientconfdir'):
        bconfagent = os.path.join(confsrv.get('clientconfdir'), client)
    else:
        click.echo(
            click.style(
                'Unable to find "clientconfdir" option, you will have to '
                'setup the agent by your own',
                fg='yellow'
            )
        )
        bconfagent = os.devnull

    if not os.path.exists(bconfcli):
        clitpl = """
mode = client
port = 4971
status_port = 4972
server = ::1
password = abcdefgh
cname = {0}
protocol = 1
pidfile = /tmp/burp.client.pid
syslog = 0
stdout = 1
progress_counter = 1
network_timeout = 72000
server_can_restore = 0
cross_all_filesystems=0
ca_burp_ca = /usr/sbin/burp_ca
ca_csr_dir = /etc/burp/CA-client
ssl_cert_ca = /etc/burp/ssl_cert_ca-client-{0}.pem
ssl_cert = /etc/burp/ssl_cert-bui-client.pem
ssl_key = /etc/burp/ssl_cert-bui-client.key
ssl_key_password = password
ssl_peer_cn = burpserver
include = /home
exclude_fs = sysfs
exclude_fs = tmpfs
nobackup = .nobackup
exclude_comp=bz2
exclude_comp=gz
""".format(client)

        if dry:
            (_, dest_bconfcli) = tempfile.mkstemp()
        with open(dest_bconfcli, 'w') as confcli:
            confcli.write(clitpl)

    parser = app.client.get_parser()

    confcli = Config(dest_bconfcli, parser, 'srv')
    confcli.set_default(dest_bconfcli)
    confcli.parse()

    if confcli.get('cname') != client:
        confcli['cname'] = client
    if confcli.get('server') != host:
        confcli['server'] = host
    if confcli.get('status_port')[0] != status_port:
        c_status_port = confcli.get_raw('status_port')
        c_status_port[0] = status_port

    if confcli.dirty:
        if dry:
            (_, dstfile) = tempfile.mkstemp()
        else:
            dstfile = bconfcli

        confcli.store(conf=bconfcli, dest=dstfile, insecure=True)
        if dry:
            before = []
            after = []
            try:
                with open(bconfcli) as fil:
                    before = fil.readlines()
            except:
                pass
            try:
                with open(dstfile) as fil:
                    after = fil.readlines()
                os.unlink(dstfile)
            except:
                pass

            if dest_bconfcli != bconfcli:
                # the file did not exist
                os.unlink(dest_bconfcli)
                before = []

            diff = difflib.unified_diff(before, after, fromfile=bconfcli, tofile='{}.new'.format(bconfcli))
            out = ''
            for line in diff:
                out += _color_diff(line)
            if out:
                click.echo_via_pager(out)

    if not os.path.exists(bconfagent):

        agenttpl = """
password = abcdefgh
"""

        if not dry:
            with open(bconfagent, 'w') as confagent:
                confagent.write(agenttpl)
        else:
            before = []
            after = ['{}\n'.format(x) for x in agenttpl.splitlines()]
            diff = difflib.unified_diff(before, after, fromfile='None', tofile=bconfagent)
            out = ''
            for line in diff:
                out += _color_diff(line)
            if out:
                click.echo_via_pager(out)

    else:
        confagent = Config(bconfagent, parser, 'cli')
        confagent.set_default(bconfagent)
        confagent.parse()

        if confagent.get('password') != confcli.get('password'):
            click.echo(
                click.style(
                    'It looks like the passwords in the {} and the {} files '
                    'mismatch. Burp-UI will not work properly until you fix '
                    'this'.format(bconfcli, bconfagent),
                    fg='yellow'
                )
            )


@app.cli.command()
@click.option('-c', '--client', default='bui',
              help='Name of the burp client that will be used by Burp-UI '
                   '(defaults to "bui")')
@click.option('-h', '--host', default='::1',
              help='Address of the status server (defaults to "::1")')
@click.option('-t', '--tips', is_flag=True,
              help='Show you some tips')
def diag(client, host, tips):
    """Check Burp-UI is correctly setup."""
    if app.vers != 2:
        click.echo(
            click.style(
                'Sorry, you can only setup the Burp 2 client',
                fg='red'
            ),
            err=True
        )
        sys.exit(1)

    if not app.standalone:
        click.echo(
            click.style(
                'Sorry, only the standalone mode is supported',
                fg='red'
            ),
            err=True
        )
        sys.exit(1)

    try:
        msg = app.load_modules(True)
    except Exception as e:
        msg = str(e)

    if msg:
        _die(msg, 'diag')

    from .misc.parser.utils import Config
    from .app import get_redis_server

    if 'Production' in app.conf.options and \
            'redis' in app.conf.options['Production']:
        try:
            # detect missing modules
            import redis as redis_client  # noqa
            import celery  # noqa
            import socket

            rhost, rport, _ = get_redis_server(app)
            ret = -1
            for res in socket.getaddrinfo(rhost, rport, socket.AF_UNSPEC, socket.SOCK_STREAM):
                if ret == 0:
                    break
                af, socktype, proto, _, sa = res
                try:
                    s = socket.socket(af, socktype, proto)
                except socket.error:
                    continue
                try:
                    ret = s.connect_ex(sa)
                except:
                    continue

            if ret != 0:
                click.echo(
                    click.style(
                        'Unable to contact the redis server, disabling it',
                        fg='yellow'
                    )
                )

        except ImportError:
            click.echo(
                click.style(
                    'Unable to activate redis & celery. Did you ran the '
                    '\'pip install "burp-ui[celery]"\' and '
                    '\'pip install "burp-ui[gunicorn-extra]"\' commands first?',
                    fg='yellow'
                )
            )

    if 'Production' in app.conf.options and \
            'database' in app.conf.options['Production']:
        try:
            from .ext.sql import db  # noqa
        except ImportError:
            click.echo(
                click.style(
                    'It looks like some dependencies are missing. Did you ran '
                    'the \'pip install "burp-ui[sql]"\' command first?',
                    fg='yellow'
                )
            )

    section = 'Burp'
    if not app.conf.section_exists(section):
        click.echo(
            click.style(
                'Section [Burp] not found, looking for the old [Burp2] section '
                'instead.',
                fg='yellow'
            )
        )
        section = 'Burp2'
        if not app.conf.section_exists(section):
            click.echo(
                click.style(
                    'No [Burp*] section found at all!',
                    fg='red'
                )
            )
            section = 'Burp'

    bconfcli = app.conf.options.get(section, {}).get('bconfcli') or \
        getattr(app.client, 'burpconfcli')
    bconfsrv = app.conf.options.get(section, {}).get('bconfsrv') or \
        getattr(app.client, 'burpconfsrv')

    try:
        app.client.status()
    except Exception as e:
        if 'Unable to spawn burp process' in str(e):
            try:
                app.client._spawn_burp(verbose=True)
            except Exception as e:
                msg = str(e)
        else:
            msg = str(e)

    if msg:
        click.echo(click.style(msg, fg='red'))
        if 'could not connect' in msg:
            click.echo(
                click.style(
                    'It looks like your burp-client can not reach your '
                    'burp-server. Please check both your \'server\' setting in '
                    'your \'{}\' file and \'status_address\' in your \'{}\' '
                    'file.'.format(bconfcli, bconfsrv),
                    fg='yellow'
                )
            )

    errors = False
    if os.path.exists(bconfcli):
        parser = app.client.get_parser()

        confcli = Config(bconfcli, parser, 'srv')
        confcli.set_default(bconfcli)
        confcli.parse()

        if confcli.get('cname') != client:
            click.echo(
                click.style(
                    'The cname of your burp client does not match: '
                    '{} != {}'.format(confcli.get('cname'), client),
                    fg='yellow'
                )
            )
            errors = True
        if confcli.get('server') != host:
            click.echo(
                click.style(
                    'The burp server address does not match: '
                    '{} != {}'.format(confcli.get('server'), host),
                    fg='yellow'
                )
            )
            errors = True

    else:
        click.echo(
            click.style(
                'No client conf file found: {} does not exist'.format(bconfcli),
                fg='red'
            ),
            err=True
        )
        errors = True

    if os.path.exists(bconfsrv):
        parser = app.client.get_parser()

        confsrv = Config(bconfsrv, parser, 'srv')
        confsrv.set_default(bconfsrv)
        confsrv.parse()

        if host not in ['::1', '127.0.0.1']:
            bind = confsrv.get('status_address')
            if (bind and bind not in [host, '::', '0.0.0.0']) or not bind:
                click.echo(
                    click.style(
                        'It looks like your burp server is not exposing it\'s '
                        'status port in a way that is reachable by Burp-UI!',
                        fg='yellow'
                    )
                )
                click.echo(
                    click.style(
                        'You may want to set the \'status_address\' setting with '
                        'either \'{}\', \'::\' or \'0.0.0.0\' in the {} file '
                        'in order to make Burp-UI work'.format(host, bconfsrv),
                        fg='blue'
                    )
                )

        max_status_children = confsrv.get('max_status_children', [-1])
        if all([x < 15 for x in max_status_children]):
            click.echo(
                click.style(
                    '\'max_status_children\' is to low, you need to set it to '
                    '15 or more. Please edit your {} file.'.format(bconfsrv),
                    fg='blue'
                )
            )
            errors = True

        restore = []
        if 'restore_client' in confsrv:
            restore = confsrv.getlist('restore_client')

        if client not in restore:
            click.echo(
                click.style(
                    'Your burp client is not listed as a \'restore_client\'. '
                    'You won\'t be able to view other clients stats!',
                    fg='yellow'
                )
            )
            errors = True

        if 'monitor_browse_cache' not in confsrv or not \
                confsrv.get('monitor_browse_cache'):
            click.echo(
                click.style(
                    'For performance reasons, it is recommended to enable the '
                    '\'monitor_browse_cache\'.',
                    fg='yellow'
                )
            )
            errors = True

        ca_client_dir = confsrv.get('ca_csr_dir')
        if ca_client_dir and not os.path.exists(ca_client_dir):
            try:
                os.makedirs(ca_client_dir)
            except IOError as exp:
                click.echo(
                    click.style(
                        'Unable to create "{}" dir: {}'.format(ca_client_dir, exp),
                        fg='yellow'
                    ),
                    err=True
                )

        if confsrv.get('clientconfdir'):
            bconfagent = os.path.join(confsrv.get('clientconfdir'), client)
        else:
            click.echo(
                click.style(
                    'Unable to find "clientconfdir" option. Something is wrong '
                    'with your setup.',
                    fg='yellow'
                )
            )
            bconfagent = 'ihopethisfiledoesnotexistbecauseitisrelatedtoburpui'

        if not os.path.exists(bconfagent) and bconfagent.startswith('/'):
            click.echo(
                click.style(
                    'Unable to find the {} file.'.format(bconfagent),
                    fg='yellow'
                )
            )
            errors = True
        else:
            confagent = Config(bconfagent, parser, 'cli')
            confagent.set_default(bconfagent)
            confagent.parse()

            if confagent.get('password') != confcli.get('password'):
                click.echo(
                    click.style(
                        'It looks like the passwords in the {} and the {} files '
                        'mismatch. Burp-UI will not work properly until you fix '
                        'this.'.format(bconfcli, bconfagent),
                        fg='yellow'
                    )
                )
    else:
        click.echo(
            click.style(
                'Unable to locate burp-server configuration: {} does not '
                'exist.'.format(bconfsrv),
                fg='red'
            ),
            err=True
        )
        errors = True

    if errors:
        if not tips:
            click.echo(
                click.style(
                    'Some errors have been found in your configuration. '
                    'Please make sure you ran this command with the right flags! '
                    '(see --help for details).'.format(sys.argv[0], sys.argv[1]),
                    fg='red'
                ),
                err=True
            )
        else:
            click.echo(
                click.style(
                    '\n'
                    'Well, if you are sure about your settings, you can run the '
                    'following command to help you setup your Burp-UI agent. '
                    '(Note, the \'--dry\' flag is here to show you the '
                    'modifications that will be applied. Once you are OK with '
                    'those, you can re-run the command without the \'--dry\' flag):',
                    fg='blue'
                )
            )
            click.echo('    > bui-manage setup_burp --host="{}" --client="{}" --dry'.format(host, client))
    else:
        click.echo(
            click.style(
                'Congratulations! It seems everything is alright. Burp-UI '
                'should run without any issue now.',
                fg='green'
            )
        )


@app.cli.command()
@click.option('-v', '--verbose', is_flag=True,
              help='Dump parts of the config (Please double check no sensitive'
              ' data leaked)')
@click.option('-l', '--load', is_flag=True,
              help='Load all configured modules for full summary')
def sysinfo(verbose, load):
    """Returns a couple of system informations to help debugging."""
    from .desc import __release__, __version__
    import platform

    msg = None
    if load:
        try:
            msg = app.load_modules(True)
        except Exception as e:
            msg = str(e)

    backend_version = app.vers
    if not app.standalone:
        backend_version = 'multi'

    colors = {
        'True': 'green',
        'False': 'red',
    }
    embedded_ws = str(app.websocket)
    available_ws = str(WS_AVAILABLE)

    click.echo('Python version:      {}.{}.{}'.format(sys.version_info[0], sys.version_info[1], sys.version_info[2]))
    click.echo('Burp-UI version:     {} ({})'.format(__version__, __release__))
    click.echo('OS:                  {}:{} ({})'.format(platform.system(), platform.release(), os.name))
    if platform.system() == 'Linux':
        click.echo('Distribution:        {} {} {}'.format(*platform.dist()))
    click.echo('Single mode:         {}'.format(app.standalone))
    click.echo('Backend version:     {}'.format(backend_version))
    click.echo('WebSocket embedded:  {}'.format(click.style(embedded_ws, fg=colors[embedded_ws])))
    click.echo('WebSocket available: {}'.format(click.style(available_ws, colors[available_ws])))
    click.echo('Config file:         {}'.format(app.config.conffile))
    if load:
        if not app.standalone and not msg:
            click.echo('Agents:')
            for agent, obj in iteritems(app.client.servers):
                client_version = server_version = 'unknown'
                try:
                    app.client.status(agent=agent)
                    client_version = app.client.get_client_version(agent=agent)
                    server_version = app.client.get_server_version(agent=agent)
                except BUIserverException:
                    pass
                alive = obj.ping()
                if alive:
                    status = click.style('ALIVE', fg='green')
                else:
                    status = click.style('DISCONNECTED', fg='red')
                click.echo(' - {} ({})'.format(agent, status))
                click.echo('   * client version: {}'.format(client_version))
                click.echo('   * server version: {}'.format(server_version))
        elif not msg:
            server_version = 'unknown'
            try:
                app.client.status()
                server_version = app.client.get_server_version()
            except BUIserverException:
                pass
            click.echo('Burp client version: {}'.format(app.client.client_version))
            click.echo('Burp server version: {}'.format(server_version))
    if verbose:
        click.echo('>>>>> Extra verbose informations:')
        click.echo(click.style(
            '!!! PLEASE MAKE SURE NO SENSITIVE DATA GET EXPOSED !!!',
            fg='red'
        ))
        sections = [
            'WebSocket',
            'Burp',
            'Production',
            'Global',
        ]
        sections.reverse()
        for section in sections:
            if section in app.config.options:
                click.echo()
                click.echo('    8<{}BEGIN[{}]'.format('-' * (67 - len(section)), section))
                for key, val in iteritems(app.config.options.get(section, {})):
                    click.echo('    {} = {}'.format(key, val))
                click.echo('    8<{}END[{}]'.format('-' * (69 - len(section)), section))

    if load and msg:
        _die(msg, 'sysinfo')



# -*- coding: utf8 -*-
"""
.. module:: burpui.decorators
    :platform: Unix
    :synopsis: Burp-UI decorators module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
import time
import datetime

from functools import wraps
from wsgiref.handlers import format_date_time
from flask_restplus.utils import unpack


def browser_cache(expires=None):
    """Add Flask cache response headers based on expires in seconds.

    If expires is None, caching will be disabled.
    Otherwise, caching headers are set to expire in now + expires seconds

    Example usage:

    ::

        @app.route('/map')
        @browser_cache(expires=60)
        def index():
            return render_template('index.html')

    """
    def cache_decorator(view):
        @wraps(view)
        def cache_func(*args, **kwargs):

            resp, code, headers = unpack(view(*args, **kwargs))
            now = datetime.datetime.now()

            headers['Last-Modified'] = format_date_time(time.mktime(now.timetuple()))

            if expires is None:
                headers['Cache-Control'] = 'no-store, no-cache, must-revalidate, post-check=0, pre-check=0, max-age=0'
                headers['Expires'] = '-1'
            else:
                expires_time = now + datetime.timedelta(seconds=expires)

                headers['Cache-Control'] = 'public'
                headers['Expires'] = format_date_time(time.mktime(expires_time.timetuple()))

            return resp, code, headers
        return cache_func
    return cache_decorator


def implement(func):
    """A decorator indicating the method is implemented.

    For the agent and the 'multi' backend, we inherit the backend interface but
    we don't really implement it because we just act as a proxy.
    But maintaining the exhaustive list of methods in several places to always
    implement the same "proxy" thing was painful so I ended up cheating to
    dynamically implement those methods thanks to the __getattribute__ magic
    function.

    But sometimes we want to implement specific things, hence this decorator
    to indicate we don't want the default "magic" implementation and use the
    custom implementation instead.
    """
    func.__ismethodimplemented__ = True
    return func



# -*- coding: utf8 -*-
"""
.. module:: burpui.ext.i18n
    :platform: Unix
    :synopsis: Burp-UI external Internationalization module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from flask import request, session
from flask_babel import Babel
from flask_login import current_user
from ..config import config
from .._compat import to_unicode

babel = Babel()

LANGUAGES = {
    'en': to_unicode('English'),
    'fr': to_unicode('Français'),
    'es': to_unicode('Español'),
    'it': to_unicode('Italiano'),
}
config['LANGUAGES'] = LANGUAGES


@babel.localeselector
def get_locale():
    locale = None
    if current_user and not current_user.is_anonymous:
        locale = getattr(current_user, 'language', None)
    elif 'language' in session:
        locale = session.get('language', None)
    if locale not in LANGUAGES:
        locale = None
    return locale or request.accept_languages.best_match(config['LANGUAGES'].keys())



# -*- coding: utf8 -*-
"""
.. module:: burpui.exceptions
    :platform: Unix
    :synopsis: Burp-UI exceptions module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
# Agent do not need "real" HTTP errors
try:
    from werkzeug.exceptions import HTTPException
    WERKZEUG = True
except ImportError:
    HTTPException = Exception
    WERKZEUG = False


class BUIserverException(HTTPException):
    """Raised in case of internal error."""
    code = 500

    def __init__(self, message="Internal Error", response=None):
        if WERKZEUG:
            HTTPException.__init__(self, message, response)
        else:
            self.description = message
            self.response = response

    def __str__(self):
        return self.description



# -*- coding: utf8 -*-
"""
.. module:: burpui._compat
    :platform: Unix
    :synopsis: Burp-UI compatibility module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
import re
import sys

try:
    import cPickle as pickle  # noqa
except ImportError:
    import pickle  # noqa

if sys.version_info[0] >= 3:
    PY3 = True
    from urllib.parse import unquote, quote, urlparse, urljoin  # noqa
    text_type = str
    string_types = (str,)

    def iterkeys(d, *args, **kwargs):
        return iter(d.keys(*args, **kwargs))

    def itervalues(d, *args, **kwargs):
        return iter(d.values(*args, **kwargs))

    def iteritems(d, *args, **kwargs):
        return iter(d.items(*args, **kwargs))

    def iterlists(d, *args, **kwargs):
        return iter(d.lists(*args, **kwargs))

    def iterlistvalues(d, *args, **kwargs):
        return iter(d.listvalues(*args, **kwargs))

else:
    PY3 = False
    from urllib import unquote, quote  # noqa
    from urlparse import urlparse, urljoin  # noqa
    text_type = unicode
    string_types = (str, unicode)

    def iterkeys(d, *args, **kwargs):
        return d.iterkeys(*args, **kwargs)

    def itervalues(d, *args, **kwargs):
        return d.itervalues(*args, **kwargs)

    def iteritems(d, *args, **kwargs):
        return d.iteritems(*args, **kwargs)

    def iterlists(d, *args, **kwargs):
        return d.iterlists(*args, **kwargs)

    def iterlistvalues(d, *args, **kwargs):
        return d.iterlistvalues(*args, **kwargs)


def to_bytes(text):
    """Transform string to bytes."""
    if isinstance(text, text_type):
        text = text.encode('utf-8')
    return text or b''


def to_unicode(input_bytes, encoding='utf-8'):
    """Decodes input_bytes to text if needed."""
    if not isinstance(input_bytes, string_types):
        input_bytes = input_bytes.decode(encoding)
    elif re.match(r'\\u[0-9a-f]{4}', input_bytes):
        input_bytes = input_bytes.decode('unicode-escape')
    return input_bytes or u''


# maps module name -> attribute name -> original item
# e.g. "time" -> "sleep" -> built-in function sleep
saved = {}


# Borrowed from gevent in order to patch json
def patch_item(module, attr, newitem, newmodule=None):
    NONE = object()
    olditem = getattr(module, attr, NONE)
    if olditem is not NONE:
        saved.setdefault(module.__name__, {}).setdefault(attr, olditem)
        if newmodule and not getattr(newmodule, 'ori_' + attr, None):
            setattr(newmodule, 'ori_' + attr, olditem)
    if not getattr(newmodule, 'ori_' + attr, None):
        setattr(module, attr, newitem)


def patch_module(name, items=None):
    toimport = items or []
    mod = __name__
    if '.' in mod:
        mod = mod.split('.')[0]
    replace_module = __import__('{}._{}'.format(mod, name), fromlist=toimport)
    module_name = name
    module = __import__(module_name)
    if items is None:
        items = getattr(replace_module, '__implements__', None)
        if items is None:
            raise AttributeError('%r does not have __implements__' % replace_module)
    for attr in items:
        patch_item(module, attr, getattr(replace_module, attr), replace_module)


def patch_json():
    try:
        import ujson  # noqa
    except ImportError:
        # ujson is not available, we won't patch anything
        return
    patch_module('json', ['dumps', 'loads'])



# -*- coding: utf8 -*-
"""
.. module:: burpui.api.custom.my_fields2
    :platform: Unix
    :synopsis: Burp-UI api custom fields module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>


"""
from flask_restplus import fields
from .my_marshalling import marshal


class Nested(fields.Nested):
    def output(self, key, obj):
        value = fields.get_value(key if self.attribute is None else self.attribute, obj)
        if value is None:
            if self.allow_null:
                return None
            elif self.default is not None:
                return self.default

        return marshal(value, self.nested)



# -*- coding: utf8 -*-
import os

from ...sessions import session_manager
from ...utils import is_uuid
from .interface import BUIhandler, BUIuser
from ..acl.interface import BUIacl

from importlib import import_module
from flask import session
from six import iteritems
from collections import OrderedDict
from flask_login import AnonymousUserMixin

ACL_METHODS = BUIacl.__abstractmethods__


class UserAuthHandler(BUIhandler):
    """See :class:`burpui.misc.auth.interface.BUIhandler`"""
    def __init__(self, app=None):
        """See :func:`burpui.misc.auth.interface.BUIhandler.__init__`

        :param app: Instance of the app we are running in
        :type app: :class:`burpui.server.BUIServer`
        """
        self.app = app
        self.users = {}
        backends = []
        self.errors = {}
        if self.app.auth and 'none' not in self.app.auth:
            me, _ = os.path.splitext(os.path.basename(__file__))
            back = self.app.auth
            for au in back:
                if au == me:
                    self.app.logger.critical('Recursive import not permitted!')
                    continue
                try:
                    (modpath, _) = __name__.rsplit('.', 1)
                    mod = import_module('.' + au, modpath)
                    obj = mod.UserHandler(self.app)
                    backends.append(obj)
                except:
                    import traceback
                    self.errors[au] = traceback.format_exc()
        for name, plugin in iteritems(self.app.plugin_manager.get_plugins_by_type('auth')):
            try:
                obj = plugin.UserHandler(self.app)
                backends.append(obj)
            except:
                import traceback
                self.errors[name] = traceback.format_exc()
        backends.sort(key=lambda x: getattr(x, 'priority', -1), reverse=True)
        if not backends:
            raise ImportError(
                'No backend found for \'{}\':\n{}'.format(self.app.auth,
                                                          self.errors)
            )
        for name, err in iteritems(self.errors):
            self.app.logger.error(
                'Unable to load module {}:\n{}'.format(repr(name), err)
            )
        self.backends = OrderedDict()
        for obj in backends:
            self.backends[obj.name] = obj

    def user(self, name=None, refresh=False):
        """See :func:`burpui.misc.auth.interface.BUIhandler.user`"""
        key = session_manager.get_session_id() or name
        if key != name and is_uuid(name) and name in self.users and \
                not session_manager.session_expired_by_id(name):
            usr = self.users[name]
            usr.id = key
            self.users[key] = usr
            del self.users[name]
            session_manager.session_import_from(name)
            session['authenticated'] = True
        if not key:
            return None
        if refresh and key in self.users:
            del self.users[key]
        if session_manager.session_managed():
            session_manager.session_expired()
        if key not in self.users:
            ret = UserHandler(self.app, self.backends, name, key)
            if not ret.name or not ret.active:
                return None
            self.users[key] = ret
            return ret
        ret = self.users[key]
        ret.refresh_session()
        self.users[key] = ret
        return ret

    def remove(self, name):
        """See :func:`burpui.misc.auth.interface.BUIhandler.remove`"""
        if name in self.users:
            del self.users[name]

    @property
    def loader(self):
        return None


class ProxyACLCall(object):
    """Class that actually calls the ACL method"""
    def __init__(self, acl, username, method):
        """
        :param acl: ACL to use
        :type acl: :class:`burpui.misc.acl.interface.BUIacl`

        :param username: username to check ACL for
        :type username: str

        :param method: Name of the method to proxify
        :type method: str
        """
        self.acl = acl
        self.username = username
        self.method = method

    def __call__(self, *args, **kwargs):
        """This is where the proxy call (and the magic) occurs"""
        # retrieve the original function prototype
        proto = getattr(BUIacl, self.method)
        args_name = list(proto.__code__.co_varnames)
        # skip self
        args_name.pop(0)
        # skip username
        args_name.pop(0)
        # we transform unnamed arguments to named ones
        # example:
        #     def my_function(toto, tata=None, titi=None):
        #
        #     x = my_function('blah', titi='blih')
        #
        # => {'toto': 'blah', 'titi': 'blih'}
        encoded_args = {
            'username': self.username
        }
        for idx, opt in enumerate(args):
            encoded_args[args_name[idx]] = opt
        encoded_args.update(kwargs)

        func = getattr(self.acl, self.method)
        return func(**encoded_args)


class ACLproxy(BUIacl):
    foreign = ACL_METHODS
    BUIacl.__abstractmethods__ = frozenset()

    def __init__(self, acl, username):
        self.acl = acl
        self.username = username

    def __getattribute__(self, name):
        # always return this value because we need it and if we don't do that
        # we'll end up with an infinite loop
        if name in ['foreign', 'acl', 'username']:
            return object.__getattribute__(self, name)
        # now we can retrieve the 'foreign' list and know if the object called
        # needs to be "proxyfied"
        if name in self.foreign:
            if self.acl:
                return ProxyACLCall(self.acl, self.username, name)
            # no ACL, assume true
            return ProxyTrue()
        return object.__getattribute__(self, name)


class ProxyTrue(object):
    def __call__(self, *args, **kwargs):
        return True


class ProxyFalse(object):
    def __call__(self, *args, **kwargs):
        return False


class ACLanon(BUIacl):
    foreign = ACL_METHODS
    BUIacl.__abstractmethods__ = frozenset()

    def __getattribute__(self, name):
        # always return this value because we need it and if we don't do that
        # we'll end up with an infinite loop
        if name == 'foreign':
            return object.__getattribute__(self, name)
        # now we can retrieve the 'foreign' list and know if the object called
        # needs to be "proxyfied"
        if name in self.foreign:
            return ProxyFalse()
        return object.__getattribute__(self, name)


class BUIanon(AnonymousUserMixin):
    _acl = ACLanon()
    name = 'Unknown'

    def login(self, passwd=None):
        return False

    @property
    def acl(self):
        return self._acl

    @property
    def is_admin(self):
        return False

    @property
    def is_moderator(self):
        return False


class UserHandler(BUIuser):
    """See :class:`burpui.misc.auth.interface.BUIuser`"""
    def __init__(self, app, backends=None, name=None, id=None):
        """
        :param app: Application context
        :type app: :class:`burpui.server.BUIServer`
        """
        self.id = id
        self.app = app
        self.active = False
        self.authenticated = session.get('authenticated', False)
        self.language = session.get('language', None)
        self.backends = backends
        self.back = None
        if not is_uuid(name):
            self.name = name
        else:
            self.name = session_manager.get_session_username() or \
                session.get('login')
        self.real = None
        if not self.name:
            return

        for _, back in iteritems(self.backends):
            user = back.user(self.name)
            if not user:
                continue
            res = user.get_id()
            if res:
                self.real = user
                self.active = True
                self.name = res
                self.back = back
                break

        self._acl = ACLproxy(self.app.acl, self.name)
        # now load the available prefs
        self._load_prefs()

    @property
    def acl(self):
        return self._acl

    @property
    def is_admin(self):
        return self.acl.is_admin()

    @property
    def is_moderator(self):
        return self.acl.is_moderator()

    def _load_prefs(self):
        session['login'] = self.name
        if self.app.config['WITH_SQL']:
            from ...models import Pref
            prefs = Pref.query.filter_by(user=self.name).all()
            for pref in prefs:
                if pref.key == 'language':
                    continue
                if hasattr(self, pref.key):
                    setattr(self, pref.key, pref.value)
                session[pref.key] = pref.value

    def refresh_session(self):
        self.authenticated = session.get('authenticated', False)
        self.language = session.get('language', None)
        self._load_prefs()

    def login(self, passwd=None):
        """See :func:`burpui.misc.auth.interface.BUIuser.login`"""
        if not self.real:
            self.authenticated = False
            for name, back in iteritems(self.backends):
                u = back.user(self.name)
                if not u:
                    continue
                res = u.get_id()
                if u.login(passwd):
                    self.authenticated = True
                    self.real = u
                    self.back = back
                    self.name = res
                    self._acl.username = res
                    break
        elif self.real:  # pragma: no cover
            if self.back and getattr(self.back, 'changed', False):
                self.real = None
                self.back = None
                return self.login(passwd)
            self.authenticated = self.real.login(passwd)
        session['authenticated'] = self.authenticated
        session['language'] = self.language
        session['login'] = self.name
        return self.authenticated



# -*- coding: utf8 -*-
"""
.. module:: burpui.server
    :platform: Unix
    :synopsis: Burp-UI server module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
import os
import re
import sys
import logging
import traceback

from .misc.auth.handler import UserAuthHandler
from .misc.acl.handler import ACLloader
from .config import config
from .plugins import PluginManager

from datetime import timedelta
from flask import Flask
from six import iteritems


G_PORT = 5000
G_BIND = u'::'
G_REFRESH = 180
G_LIVEREFRESH = 5
G_IGNORE_LABELS = ["color:.*"]
G_FORMAT_LABELS = ["s/^os:\s*//"]
G_DEFAULT_STRIP = 0
G_SSL = False
G_SINGLE = True
G_SSLCERT = u''
G_SSLKEY = u''
G_VERSION = 2
G_AUTH = [u'basic']
G_ACL = u'none'
G_STORAGE = u''
G_CACHE = u''
G_SESSION = u''
G_REDIS = u''
G_LIMITER = False
G_RATIO = u'60/minute'
G_CELERY = False
G_SCOOKIE = True
G_DEMO = False
G_DSN = u''
G_PIWIK_URL = u''
G_PIWIK_SCRIPT = u'piwik.php'
G_PIWIK_ID = 0
G_APPSECRET = u'random'
G_COOKIETIME = 14
G_SESSIONTIME = 5
G_DATABASE = u''
G_PREFIX = u''
G_PLUGINS = []
G_NO_SERVER_RESTORE = False
G_WS_ENABLED = True
G_WS_EMBEDDED = False
G_WS_BROKER = u'redis'
G_WS_URL = u''
G_WS_DEBUG = False


class BUIServer(Flask):
    """
    The :class:`burpui.server.BUIServer` class provides the ``Burp-UI`` server.
    """
    gunicorn = False

    defaults = {
        'Global': {
            'port': G_PORT,
            'bind': G_BIND,
            'ssl': G_SSL,
            'standalone': G_SINGLE,
            'single': G_SINGLE,
            'sslcert': G_SSLCERT,
            'sslkey': G_SSLKEY,
            'version': G_VERSION,
            'auth': G_AUTH,
            'acl': G_ACL,
            'prefix': G_PREFIX,
            'plugins': G_PLUGINS,
            'demo': G_DEMO,
            'dsn': G_DSN,
            'piwik_url': G_PIWIK_URL,
            'piwik_script': G_PIWIK_SCRIPT,
            'piwik_id': G_PIWIK_ID,
        },
        'UI': {
            'refresh': G_REFRESH,
            'liverefresh': G_LIVEREFRESH,
            'ignore_labels': G_IGNORE_LABELS,
            'format_labels': G_FORMAT_LABELS,
            'default_strip': G_DEFAULT_STRIP,
        },
        'Security': {
            'scookie': G_SCOOKIE,
            'appsecret': G_APPSECRET,
            'cookietime': G_COOKIETIME,
            'sessiontime': G_SESSIONTIME,
        },
        'Production': {
            'storage': G_STORAGE,
            'session': G_SESSION,
            'cache': G_CACHE,
            'redis': G_REDIS,
            'celery': G_CELERY,
            'database': G_DATABASE,
            'limiter': G_LIMITER,
            'ratio': G_RATIO,
        },
        'WebSocket': {
            'enabled': G_WS_ENABLED,
            'embedded': G_WS_EMBEDDED,
            'broker': G_WS_BROKER,
            'url': G_WS_URL,
            'debug': G_WS_DEBUG,
        },
        'Experimental': {
            'noserverrestore': G_NO_SERVER_RESTORE,
        }
    }

    def __init__(self):
        """The :class:`burpui.server.BUIServer` class provides the ``Burp-UI``
        server.

        :param app: The Flask application to launch
        """
        super(BUIServer, self).__init__('burpui')
        self.init = False
        # We cannot override the Flask's logger so we use our own
        self._logger = logging.getLogger('burp-ui')
        self._logger.disabled = True
        self.conf = config
        # switch the flask config with our magic config object
        self.conf.update(self.config)
        self.config = self.conf

    def enable_logger(self, enable=True):
        """Enable or disable the logger"""
        self._logger.disabled = not enable

    @property
    def logger(self):
        """
        :rtype: :class:`logging.Logger`
        """
        return self._logger

    def setup(self, conf=None, unittest=False, cli=False):
        """The :func:`burpui.server.BUIServer.setup` functions is used to setup
        the whole server by parsing the configuration file and loading the
        different backends.

        :param conf: Path to a configuration file
        :type conf: str

        :param unittest: Whether we are unittesting or not (used to avoid burp2
                         strict requirements checks)
        :type unittest: bool

        :param cli: Whether we are in cli mode or not
        :type cli: bool
        """
        self.sslcontext = None
        if not conf:
            conf = self.config['CFG']

        if not conf:
            raise IOError('No configuration file found')

        # Raise exception if errors are encountered during parsing
        self.conf.parse(conf, True, self.defaults)
        self.conf.default_section('Global')

        self.port = self.config['BUI_PORT'] = self.conf.safe_get(
            'port',
            'integer'
        )
        self.demo = self.config['BUI_DEMO'] = self.conf.safe_get(
            'demo',
            'boolean'
        )
        self.config['BUI_DSN'] = self.conf.safe_get('dsn')
        self.config['BUI_PIWIK_URL'] = self.conf.safe_get('piwik_url')
        self.config['BUI_PIWIK_SCRIPT'] = self.conf.safe_get('piwik_script')
        self.config['BUI_PIWIK_ID'] = self.conf.safe_get('piwik_id', 'integer')
        self.bind = self.config['BUI_BIND'] = self.conf.safe_get('bind')
        version = self.conf.safe_get('version', 'integer')
        if unittest and version != 1:
            version = 1
        self.vers = self.config['BUI_VERS'] = version
        self.ssl = self.config['BUI_SSL'] = self.conf.safe_get(
            'ssl',
            'boolean'
        )
        # option standalone has been renamed for less confusion
        key = 'standalone' if 'standalone' in \
            self.conf.conf.get(self.conf.section, {}) else 'single'
        if key == 'standalone':
            # TODO: remove the compatibility in v0.7.0
            self.logger.warning(
                'The "standalone" option is DEPRECATED and has been replaced '
                'by the "single" option. Please update your conf before we '
                'remove the compatibility in v0.7.0'
            )
        self.standalone = self.config['STANDALONE'] = self.conf.safe_get(
            key,
            'boolean'
        )
        self.sslcert = self.config['BUI_SSLCERT'] = self.conf.safe_get(
            'sslcert'
        )
        self.sslkey = self.config['BUI_SSLKEY'] = self.conf.safe_get(
            'sslkey'
        )
        self.prefix = self.config['BUI_PREFIX'] = self.conf.safe_get(
            'prefix'
        )
        if self.prefix and not self.prefix.startswith('/'):
            if self.prefix.lower() != 'none':
                self.logger.warning("'prefix' must start with a '/'!")
            self.prefix = self.config['BUI_PREFIX'] = ''

        self.plugins = self.config['BUI_PLUGINS'] = self.conf.safe_get(
            'plugins',
            'string_lower_list'
        )
        if len(self.plugins) == 1 and self.plugins[0] == 'none':
            self.plugins = self.config['BUI_PLUGINS'] = []

        self.auth = self.config['BUI_AUTH'] = self.conf.safe_get(
            'auth',
            'string_lower_list'
        )

        # UI options
        self.config['REFRESH'] = self.conf.safe_get(
            'refresh',
            'integer',
            'UI'
        )
        self.config['LIVEREFRESH'] = self.conf.safe_get(
            'liverefresh',
            'integer',
            'UI'
        )
        self.config['DEFAULT_STRIP'] = self.conf.safe_get(
            'default_strip',
            'integer',
            'UI'
        )
        self.ignore_labels = self.conf.safe_get(
            'ignore_labels',
            'force_list',
            'UI'
        )
        format_labels = self.conf.safe_get(
            'format_labels',
            'force_list',
            'UI'
        )
        self.format_labels = []
        for format_label in format_labels:
            search = re.search(r'^s(?P<separator>.)(?P<regex>.*?)(?P=separator)(?P<replace>.*?)(?P=separator)$', format_label)
            if search:
                self.format_labels.append((search.group('regex'), search.group('replace')))

        # Production options
        self.storage = self.config['BUI_STORAGE'] = self.conf.safe_get(
            'storage',
            section='Production'
        )
        self.cache_db = self.config['BUI_CACHE_DB'] = self.conf.safe_get(
            'cache',
            section='Production'
        )
        self.session_db = self.config['BUI_SESSION_DB'] = self.conf.safe_get(
            'session',
            section='Production'
        )
        self.redis = self.config['BUI_REDIS'] = self.conf.safe_get(
            'redis',
            section='Production'
        )
        self.limiter = self.config['BUI_LIMITER'] = self.conf.safe_get(
            'limiter',
            'boolean_or_string',
            section='Production'
        )
        if isinstance(self.limiter, bool) and not self.limiter:
            self.limiter = self.config['BUI_LIMITER'] = 'none'
        self.ratio = self.config['BUI_RATIO'] = self.conf.safe_get(
            'ratio',
            section='Production'
        )
        self.use_celery = self.config['BUI_CELERY'] = self.conf.safe_get(
            'celery',
            'boolean_or_string',
            section='Production'
        )
        self.database = self.config['SQLALCHEMY_DATABASE_URI'] = \
            self.conf.safe_get(
                'database',
                'boolean_or_string',
                section='Production'
        )
        self.config['WITH_LIMIT'] = False
        if isinstance(self.database, bool):
            self.config['WITH_SQL'] = self.database
        else:
            self.config['WITH_SQL'] = self.database and \
                self.database.lower() != 'none'
        if isinstance(self.use_celery, bool):
            self.config['WITH_CELERY'] = self.use_celery
        else:
            self.config['WITH_CELERY'] = self.use_celery and \
                self.use_celery.lower() != 'none'

        # WebSocket options
        self.ws_enabled = self.config['WS_ENABLED'] = self.conf.safe_get(
            'enabled',
            'boolean',
            section='WebSocket'
        )
        self.websocket = self.config['WITH_WS'] = self.conf.safe_get(
            'embedded',
            'boolean',
            section='WebSocket'
        )
        self.ws_broker = self.config['BUI_WS_BROKER'] = self.conf.safe_get(
            'broker',
            'boolean_or_string',
            section='WebSocket'
        )
        self.config['WS_DEBUG'] = self.conf.safe_get(
            'debug',
            'boolean',
            section='WebSocket'
        )
        self.config['WS_URL'] = self.conf.safe_get(
            'url',
            section='WebSocket'
        )
        if self.config.get('WS_URL', '').lower() == 'none' or self.websocket:
            self.config['WS_URL'] = None

        # Experimental options
        self.noserverrestore = self.conf.safe_get(
            'noserverrestore',
            'boolean',
            section='Experimental'
        )

        # Security options
        self.scookie = self.config['BUI_SCOOKIE'] = self.conf.safe_get(
            'scookie',
            'boolean',
            section='Security'
        )
        self.config['SECRET_KEY'] = self.conf.safe_get(
            'appsecret',
            section='Security'
        )
        days = self.conf.safe_get('cookietime', 'integer', section='Security') \
            or G_COOKIETIME
        self.config['REMEMBER_COOKIE_DURATION'] = \
            self.config['PERMANENT_SESSION_LIFETIME'] = timedelta(
                days=days
        )
        self.config['REMEMBER_COOKIE_NAME'] = 'remember_token'
        days = self.conf.safe_get(
            'sessiontime',
            'integer',
            section='Security'
        ) or G_SESSIONTIME
        self.config['SESSION_INACTIVE'] = timedelta(days=days)

        self.logger.info('burp version: {}'.format(self.vers))
        self.logger.info('listen port: {}'.format(self.port))
        self.logger.info('bind addr: {}'.format(self.bind))
        self.logger.info('use ssl: {}'.format(self.ssl))
        self.logger.info('standalone: {}'.format(self.standalone))
        self.logger.info('sslcert: {}'.format(self.sslcert))
        self.logger.info('sslkey: {}'.format(self.sslkey))
        self.logger.info('prefix: {}'.format(self.prefix))
        self.logger.info('secure cookie: {}'.format(self.scookie))
        self.logger.info(
            'cookietime: {}'.format(self.config['REMEMBER_COOKIE_DURATION'])
        )
        self.logger.info(
            'session inactive: {}'.format(self.config['SESSION_INACTIVE'])
        )
        self.logger.info('refresh: {}'.format(self.config['REFRESH']))
        self.logger.info('liverefresh: {}'.format(self.config['LIVEREFRESH']))
        self.logger.info('auth: {}'.format(self.auth))
        self.logger.info('celery: {}'.format(self.use_celery))
        self.logger.info('redis: {}'.format(self.redis))
        self.logger.info('limiter: {}'.format(self.limiter))
        self.logger.info('database: {}'.format(self.database))
        self.logger.info('with SQL: {}'.format(self.config['WITH_SQL']))
        self.logger.info('with Celery: {}'.format(self.config['WITH_CELERY']))
        self.logger.info('with WebSocket: {}'.format(self.config['WITH_WS']))
        self.logger.info('demo: {}'.format(self.config['BUI_DEMO']))

        self.init = True
        if not cli:
            self.load_modules()

    def load_modules(self, strict=False):
        """Load the extensions"""
        self.plugin_manager = PluginManager(self, self.plugins)
        if self.plugins:
            self.plugin_manager.load_all()

        if self.auth and 'none' not in self.auth:
            try:
                self.uhandler = UserAuthHandler(self)
                for back, err in iteritems(self.uhandler.errors):
                    self.logger.critical(
                        'Unable to load \'{}\' authentication backend:\n{}'
                        .format(back, err)
                    )
            except ImportError as e:
                self.logger.critical(
                    'Import Exception, module \'{0}\': {1}'.format(
                        self.auth,
                        str(e)
                    )
                )
                raise e
            self.acl_engine = self.config['BUI_ACL'] = self.conf.safe_get(
                'acl',
                'string_lower_list'
            )
            self.config['LOGIN_DISABLED'] = False
        else:
            self.config['LOGIN_DISABLED'] = True
            # No login => no ACL
            self.acl_engine = self.config['BUI_ACL'] = ['none']
            self.auth = self.config['BUI_AUTH'] = 'none'

        if self.acl_engine and 'none' not in self.acl_engine:
            try:
                self.acl_handler = ACLloader(self)
            except Exception as e:
                self.logger.critical(
                    'Import Exception, module \'{0}\': {1}'.format(
                        self.acl_engine,
                        str(e)
                    )
                )
                raise e
        else:
            self.acl_handler = False

        self.logger.info('acl: {}'.format(self.acl_engine))

        if self.standalone:
            module = 'burpui.misc.backend.burp{0}'.format(self.vers)
        else:
            module = 'burpui.misc.backend.multi'

        # This is used for development purpose only
        from .misc.backend.burp1 import Burp as BurpGeneric
        self.client = BurpGeneric(dummy=True)
        self.strict = strict
        try:
            # Try to load submodules from our current environment
            # first
            sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))
            mod = __import__(module, fromlist=['Burp'])
            Client = mod.Burp
            self.client = Client(self, conf=self.conf)
        except Exception as e:
            msg = 'Failed loading backend for Burp version {0}: {1}'.format(
                self.vers,
                str(e)
            )
            if strict:
                self.logger.critical(traceback.format_exc())
                self.logger.critical(msg)
                sys.exit(2)
            else:
                raise Exception(msg)

    @property
    def acl(self):
        """ACL module

        :returns: :class:`burpui.misc.acl.interface.BUIacl`
        """
        if self.acl_engine and 'none' not in self.acl_engine:
            # refresh acl to detect config changes
            from .misc.acl.interface import BUIacl  # noqa
            acl = self.acl_handler.acl  # type: BUIacl
            return acl
        return None

    def get_send_file_max_age(self, name):
        """Provides default cache_timeout for the send_file() functions."""
        if name:
            lname = name.lower()
            extensions = ['js', 'css', 'woff']
            for ext in extensions:
                if lname.endswith('.{}'.format(ext)):
                    return 3600 * 24 * 30  # 30 days
        return Flask.get_send_file_max_age(self, name)

    def manual_run(self):
        """The :func:`burpui.server.BUIServer.manual_run` functions is used to
        actually launch the ``Burp-UI`` server.

        .. deprecated:: 0.4.0
           You should now run the Flask's run command.
        """
        if not self.init:
            self.setup()

        if self.ssl:
            self.sslcontext = (self.sslcert, self.sslkey)

        if self.sslcontext:
            self.config['SSL'] = True
            self.run(
                host=self.bind,
                port=self.port,
                debug=self.config['DEBUG'],
                ssl_context=self.sslcontext
            )
        else:
            self.run(host=self.bind, port=self.port, debug=self.config['DEBUG'])



# -*- coding: utf8 -*-
"""
.. module:: burpui.api.admin
    :platform: Unix
    :synopsis: Burp-UI admin api module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from . import api
from ..server import BUIServer  # noqa
from ..sessions import session_manager
from ..misc.acl.meta import meta_grants
from ..utils import NOTIF_OK
from .custom import fields, Resource
#  from ..exceptions import BUIserverException

from six import iteritems
from flask import current_app
from flask_login import current_user
from flask_babel import gettext

import json

bui = current_app  # type: BUIServer
ns = api.namespace('admin', 'Admin methods')

user_fields = ns.model('Users', {
    'id': fields.String(required=True, description='User id'),
    'name': fields.String(required=True, description='User name'),
    'backend': fields.String(required=True, description='Backend name'),
})
grant_fields = ns.model('Grants', {
    'id': fields.String(required=True, description='Grant id'),
    'grant': fields.String(required=True, description='Grant content'),
    'backend': fields.String(required=True, description='Backend name'),
})
group_fields = ns.model('Groups', {
    'id': fields.String(required=True, description='Group id'),
    'grant': fields.String(required=True, description='Group grant content'),
    'members': fields.List(fields.String, required=True, description='Group members'),
    'backend': fields.String(required=True, description='Backend name'),
})
groups_fields = ns.model('GroupsFields', {
    'name': fields.String(required=True, description='Group name'),
    'inherit': fields.List(fields.String, required=False, description='This group is inherited by'),
})
groups_list_fields = ns.model('GroupsList', {
    'groups': fields.List(fields.Nested(groups_fields), required=True, description='Groups list'),
})
group_members_fields = ns.model('GroupMembers', {
    'members': fields.List(fields.String, required=True, description='Group members'),
    'grant': fields.String(required=True, description='Group grant content'),
})
is_moderator_fields = ns.model('IsModerator', {
    'moderator': fields.Boolean(required=True, description='Is the member a moderator'),
    'inherit': fields.List(fields.String, required=False, description='What provides this grant if inherited'),
})
moderator_members_fields = ns.model('ModeratorMembers', {
    'members': fields.List(fields.String, required=True, description='Moderator members'),
    'grant': fields.String(required=True, description='Moderator grant content'),
})
moderators_fields = ns.model('Moderators', {
    'members': fields.List(fields.String, required=True, description='Moderator members'),
    'grant': fields.String(required=True, description='Moderator grant content'),
    'backend': fields.String(required=True, description='Backend name'),
})
is_admin_fields = ns.model('IsAdmin', {
    'admin': fields.Boolean(required=True, description='Is the member an admin'),
    'inherit': fields.List(fields.String, required=False, description='What provides this grant if inherited'),
})
admin_members_fields = ns.model('AdminMembers', {
    'members': fields.List(fields.String, required=True, description='Admin members'),
})
admins_fields = ns.model('Admins', {
    'members': fields.List(fields.String, required=True, description='Admin members'),
    'backend': fields.String(required=True, description='Backend name'),
})
session_fields = ns.model('Sessions', {
    'uuid': fields.String(description='Session id'),
    'ip': fields.String(description='IP address'),
    'ua': fields.String(description='User-Agent'),
    'permanent': fields.Boolean(description='Remember cookie'),
    'api': fields.Boolean(description='API login'),
    'expire': fields.DateTime(description='Expiration date'),
    'timestamp': fields.DateTime(description='Last seen'),
    'current': fields.Boolean(description='Is current session', default=False)
})
acl_backend_fields = ns.model('AclBackends', {
    'name': fields.String(required=True, description='Backend name'),
    'description': fields.String(required=True, description='Backend description'),
    'type': fields.String(required=False, description='Backend type'),
    'priority': fields.Integer(required=False, description='Backend priority'),
    'add_grant': fields.Boolean(required=False, default=False, description='Support grant creation'),
    'mod_grant': fields.Boolean(required=False, default=False, description='Support grant edition'),
    'del_grant': fields.Boolean(required=False, default=False, description='Support grant deletion'),
    'add_group': fields.Boolean(required=False, default=False, description='Support group creation'),
    'mod_group': fields.Boolean(required=False, default=False, description='Support group edition'),
    'del_group': fields.Boolean(required=False, default=False, description='Support group deletion'),
    'add_group_member': fields.Boolean(required=False, default=False, description='Support group member addition'),
    'del_group_member': fields.Boolean(required=False, default=False, description='Support group member deletion'),
    'add_moderator': fields.Boolean(required=False, default=False, description='Support moderator creation'),
    'mod_moderator': fields.Boolean(required=False, default=False, description='Support moderator edition'),
    'del_moderator': fields.Boolean(required=False, default=False, description='Support moderator deletion'),
    'add_admin': fields.Boolean(required=False, default=False, description='Support admin creation'),
    'del_admin': fields.Boolean(required=False, default=False, description='Support admin deletion'),
})
auth_backend_fields = ns.model('Backends', {
    'name': fields.String(required=True, description='Backend name'),
    'description': fields.String(required=True, description='Backend description'),
    'type': fields.String(required=False, description='Backend type'),
    'priority': fields.Integer(required=False, description='Backend priority'),
    'add': fields.Boolean(required=False, default=False, description='Support user creation'),
    'mod': fields.Boolean(required=False, default=False, description='Support user edition'),
    'del': fields.Boolean(required=False, default=False, description='Support user deletion'),
})


@ns.route('/me', endpoint='admin_me')
class AdminMe(Resource):
    """The :class:`burpui.api.admin.AdminMe` resource allows you to
    retrieve informations about your currently logged in user.

    This resource is part of the :mod:`burpui.api.admin` module.
    """

    @ns.marshal_with(user_fields)
    def get(self):
        """Returns the current user informations

        **GET** method provided by the webservice.

        :returns: User
        """
        ret = getattr(current_user, 'real', current_user)
        return ret


@ns.route('/acl/isAdmin/<member>',
          '/acl/<backend>/isAdmin/<member>',
          endpoint='acl_is_admin')
@ns.doc(
    params={
        'member': 'Username',
    }
)
class AclIsAdmin(Resource):
    """The :class:`burpui.api.admin.AclIsAdmin` resources allows you to check
    if a given member is admin or not.

    This resource is part of the :mod:`burpui.api.admin` module.
    """

    @api.acl_admin_or_moderator_required(message="Not allowed to view admins list")
    @ns.marshal_with(is_admin_fields)
    @ns.doc(
        responses={
            403: 'Insufficient permissions',
            404: 'No backend found',
        },
    )
    def get(self, member, backend=None):
        """Checks if a given member is admin"""
        if not backend:
            (ret, inh) = meta_grants.is_admin(member)
            return {'admin': ret, 'inherit': inh}
        try:
            handler = getattr(bui, 'acl_handler')
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        if backend not in handler.backends:
            self.abort(404, "No acl backend '{}' found".format(backend))
        loader = handler.backends[backend]
        return {'admin': member in loader.admins}


@ns.route('/acl/admins',
          '/acl/<backend>/admins',
          endpoint='acl_admins')
@ns.doc(
    params={
        'backend': 'Backend name',
    }
)
class AclAdminss(Resource):
    """The :class:`burpui.api.admin.AclAdminss` resource allows you to
    retrieve a list of admins.

    This resource is part of the :mod:`burpui.api.admin` module.
    """

    @api.acl_admin_or_moderator_required(message="Not allowed to view groups list")
    @ns.marshal_list_with(admins_fields, code=200, description='Success')
    @ns.doc(
        responses={
            403: 'Insufficient permissions',
            404: 'No backend found',
        },
    )
    def get(self, name=None, backend=None):
        """Returns a list of admins

        **GET** method provided by the webservice.

        :returns: Moderators
        """
        try:
            handler = getattr(bui, 'acl_handler')
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        ret = []
        for _, loader in iteritems(handler.backends):
            append = {
                'members': loader.admins,
                'backend': loader.name
            }
            if (backend and backend == append['backend']) or backend is None:
                return [append]
            ret.append(append)
        return ret


@ns.route('/acl/admin',
          '/acl/<backend>/admin',
          '/acl/<backend>/admin/<member>',
          endpoint='acl_admin')
@ns.doc(
    params={
        'backend': 'ACL backend',
        'member': 'Admin member',
    }
)
class AclAdmin(Resource):
    """The :class:`burpui.api.admin.AclAdmins` resource allows you to
    retrieve a list of admins and add/delete them if your
    acl backend support those actions.

    This resource is part of the :mod:`burpui.api.admin` module.
    """
    parser = ns.parser()
    parser.add_argument('memberNames', required=False, help='Admin members', action='append')
    parser.add_argument('backendName', required=False, help='Backend name')

    @api.acl_admin_or_moderator_required(message="Not allowed to view admins list")
    @ns.marshal_with(admin_members_fields, code=200, description='Success')
    @ns.doc(
        responses={
            403: 'Insufficient permissions',
            404: 'No backend found',
        },
    )
    def get(self, backend):
        """Returns a list of admin users

        **GET** method provided by the webservice.

        :returns: Members
        """
        try:
            handler = getattr(bui, 'acl_handler')
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        if backend not in handler.backends:
            self.abort(404, "No acl backend '{}' found".format(backend))
        ret = {}
        loader = handler.backends[backend]
        ret = {
            'members': loader.admins
        }
        return ret

    @api.disabled_on_demo()
    @api.acl_admin_required(message="Not allowed to add admin members")
    @ns.expect(parser)
    @ns.doc(
        responses={
            403: 'Insufficient permissions',
            404: 'No backend found',
        },
    )
    def put(self, backend=None, member=None):
        """Add a member as admin

        **PUT** method provided by the webservice.
        """
        args = self.parser.parse_args()
        backend = backend or args['backendName']
        try:
            handler = getattr(bui, 'acl_handler')
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        if backend not in handler.backends:
            self.abort(404, "No acl backend '{}' found".format(backend))

        loader = handler.backends[backend]

        members = [member] if member else (args['memberNames'] or [])

        if loader.add_admin is False:
            self.abort(
                500,
                "The '{}' backend does not support moderator member addition"
                "".format(backend)
            )

        ret = []
        status = 200

        for member in members:
            success, message, code = loader.add_admin(
                member
            )
            status = 201 if success else 200
            ret.append([code, message])
        return ret, status

    @api.disabled_on_demo()
    @api.acl_admin_required(message="Not allowed to remove admin members")
    @ns.expect(parser)
    @ns.doc(
        responses={
            403: 'Insufficient permissions',
            404: 'No backend found',
        },
    )
    def delete(self, backend=None, member=None):
        """Remove an admin member

        **DELETE** method provided by the webservice.
        """
        args = self.parser.parse_args()
        backend = backend or args['backendName']
        try:
            handler = getattr(bui, 'acl_handler')
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        if backend not in handler.backends:
            self.abort(40422, "No acl backend '{}' found".format(backend))

        loader = handler.backends[backend]

        members = [member] if member else (args['memberNames'] or [])

        if loader.del_admin is False:
            self.abort(
                500,
                "The '{}' backend does not support admin member deletion"
                "".format(backend)
            )

        ret = []
        status = 200

        for member in members:
            success, message, code = loader.del_admin(
                member
            )
            status = 201 if success else 200
            ret.append([code, message])
        return ret, status


@ns.route('/acl/isModerator/<member>',
          '/acl/<backend>/isModerator/<member>',
          endpoint='acl_is_moderator')
@ns.doc(
    params={
        'member': 'Username',
    }
)
class AclIsModerator(Resource):
    """The :class:`burpui.api.admin.AclIsModerator` resources allows you to check
    if a given member is moderator or not.

    This resource is part of the :mod:`burpui.api.admin` module.
    """

    @api.acl_admin_or_moderator_required(message="Not allowed to view admins list")
    @ns.marshal_with(is_moderator_fields)
    @ns.doc(
        responses={
            403: 'Insufficient permissions',
            404: 'No backend found',
        },
    )
    def get(self, member, backend=None):
        """Checks if a given member is moderator"""
        if not backend:
            (ret, inh) = meta_grants.is_moderator(member)
            return {'moderator': ret, 'inherit': inh}
        try:
            handler = getattr(bui, 'acl_handler')
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        if backend not in handler.backends:
            self.abort(404, "No acl backend '{}' found".format(backend))
        loader = handler.backends[backend]
        return {'moderator': member in loader.moderators}


@ns.route('/acl/moderators',
          '/acl/<backend>/moderators',
          endpoint='acl_moderators')
@ns.doc(
    params={
        'backend': 'Backend name',
    }
)
class AclModerators(Resource):
    """The :class:`burpui.api.admin.AclModerators` resource allows you to
    retrieve a list of moderators.

    This resource is part of the :mod:`burpui.api.admin` module.
    """

    @api.acl_admin_or_moderator_required(message="Not allowed to view groups list")
    @ns.marshal_list_with(moderators_fields, code=200, description='Success')
    @ns.doc(
        responses={
            403: 'Insufficient permissions',
            404: 'No backend found',
        },
    )
    def get(self, name=None, backend=None):
        """Returns a list of moderators

        **GET** method provided by the webservice.

        :returns: Moderators
        """
        try:
            handler = getattr(bui, 'acl_handler')
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        ret = []
        for _, loader in iteritems(handler.backends):
            append = {
                'grant': loader.moderator,
                'members': loader.moderators,
                'backend': loader.name
            }
            if (backend and backend == append['backend']) or backend is None:
                return [append]
            ret.append(append)
        return ret


@ns.route('/acl/moderator',
          '/acl/<backend>/moderator',
          '/acl/<backend>/moderator/<member>',
          endpoint='acl_moderator')
@ns.doc(
    params={
        'backend': 'ACL backend',
        'member': 'Moderator member',
    }
)
class AclModerator(Resource):
    """The :class:`burpui.api.admin.AclModerator` resource allows you to
    retrieve a list of moderators and add/delete them if your
    acl backend support those actions.

    This resource is part of the :mod:`burpui.api.admin` module.
    """
    parser = ns.parser()
    parser.add_argument('memberNames', required=False, help='Moderator members', action='append')
    parser.add_argument('backendName', required=False, help='Backend name')

    parser_mod = ns.parser()
    parser_mod.add_argument('grant', required=False, help='Moderator grants')

    @api.acl_admin_or_moderator_required(message="Not allowed to view moderators list")
    @ns.marshal_with(moderator_members_fields, code=200, description='Success')
    @ns.doc(
        responses={
            403: 'Insufficient permissions',
            404: 'No backend found',
        },
    )
    def get(self, backend):
        """Returns a list of moderator users

        **GET** method provided by the webservice.

        :returns: Members
        """
        try:
            handler = getattr(bui, 'acl_handler')
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        if backend not in handler.backends:
            self.abort(404, "No acl backend '{}' found".format(backend))
        ret = {}
        loader = handler.backends[backend]
        ret = {
            'members': loader.moderators,
            'grant': loader.moderator
        }
        return ret

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(message="Not allowed to add moderator members")
    @ns.expect(parser)
    @ns.doc(
        responses={
            403: 'Insufficient permissions',
            404: 'No backend found',
        },
    )
    def put(self, backend=None, member=None):
        """Add a member as moderator

        **PUT** method provided by the webservice.
        """
        args = self.parser.parse_args()
        backend = backend or args['backendName']
        try:
            handler = getattr(bui, 'acl_handler')
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        if backend not in handler.backends:
            self.abort(404, "No acl backend '{}' found".format(backend))

        loader = handler.backends[backend]

        members = [member] if member else (args['memberNames'] or [])

        if loader.add_moderator is False:
            self.abort(
                500,
                "The '{}' backend does not support moderator member addition"
                "".format(backend)
            )

        ret = []
        status = 200
        for member in members:
            success, message, code = loader.add_moderator(
                member
            )
            ret.append([code, message])
            status = 201 if success else 200
        return ret, status

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(message="Not allowed to remove moderator members")
    @ns.expect(parser)
    @ns.doc(
        responses={
            403: 'Insufficient permissions',
            404: 'No backend found',
        },
    )
    def delete(self, backend=None, member=None):
        """Remove a moderator member

        **DELETE** method provided by the webservice.
        """
        args = self.parser.parse_args()
        backend = backend or args['backendName']
        try:
            handler = getattr(bui, 'acl_handler')
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        if backend not in handler.backends:
            self.abort(404, "No acl backend '{}' found".format(backend))

        loader = handler.backends[backend]

        members = [member] if member else (args['memberNames'] or [])

        if loader.del_moderator is False:
            self.abort(
                500,
                "The '{}' backend does not support moderator member deletion"
                "".format(backend)
            )

        ret = []
        status = 200
        for member in members:
            success, message, code = loader.del_moderator(
                member
            )
            ret.append([code, message])
            status = 201 if success else 200
        return ret, status

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(message="Not allowed to update moderator grants")
    @ns.expect(parser_mod)
    @ns.doc(
        responses={
            403: 'Insufficient permissions',
            404: 'No backend found',
        },
    )
    def post(self, backend):
        """Update moderator grants

        **POST** method provided by the webservice.
        """
        args = self.parser_mod.parse_args()
        try:
            handler = getattr(bui, 'acl_handler')
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        if backend not in handler.backends:
            self.abort(404, "No acl backend '{}' found".format(backend))

        loader = handler.backends[backend]

        grants = args['grant']

        if loader.mod_moderator is False:
            self.abort(
                500,
                "The '{}' backend does not support moderator grants edition"
                "".format(backend)
            )

        success, message, code = loader.mod_moderator(
            grants
        )
        status = 201 if success else 200
        return [[code, message]], status


@ns.route('/acl/group',
          '/acl/<backend>/group/<name>',
          '/acl/<backend>/group/<name>/<member>',
          endpoint='acl_group_members')
@ns.doc(
    params={
        'name': 'Group name',
        'backend': 'ACL backend',
        'member': 'Group member',
    }
)
class AclGroup(Resource):
    """The :class:`burpui.api.admin.AclGroup` resource allows you to
    retrieve a list of members in a given group and add/delete them if your
    acl backend support those actions.

    This resource is part of the :mod:`burpui.api.admin` module.
    """
    parser = ns.parser()
    parser.add_argument('memberNames', required=False, help='Group members', action='append')
    parser.add_argument('groupName', required=False, help='Group name')
    parser.add_argument('backendName', required=False, help='Backend name')

    @api.acl_admin_or_moderator_required(message="Not allowed to view groups list")
    @ns.marshal_with(group_members_fields, code=200, description='Success')
    @ns.doc(
        responses={
            403: 'Insufficient permissions',
            404: 'No backend found',
        },
    )
    def get(self, name, backend):
        """Returns a list of users in a giver group

        **GET** method provided by the webservice.

        :returns: Members
        """
        try:
            handler = getattr(bui, 'acl_handler')
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        if backend not in handler.backends:
            self.abort(404, "No acl backend '{}' found".format(backend))
        ret = {}
        loader = handler.backends[backend]
        groups = loader.groups
        gname = '@{}'.format(name)
        if groups and gname in groups:
            ret = {
                'members': groups[gname].get('members', []),
                'grant': groups[gname].get('grants', '')
            }
        return ret

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(message="Not allowed to add member in group")
    @ns.expect(parser)
    @ns.doc(
        responses={
            403: 'Insufficient permissions',
            404: 'No backend found',
        },
    )
    def put(self, name=None, backend=None, member=None):
        """Add a member in a given group

        **PUT** method provided by the webservice.
        """
        args = self.parser.parse_args()
        name = name or args['groupName']
        backend = backend or args['backendName']
        try:
            handler = getattr(bui, 'acl_handler')
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        if backend not in handler.backends:
            self.abort(404, "No acl backend '{}' found".format(backend))

        loader = handler.backends[backend]

        members = [member] if member else (args['memberNames'] or [])

        if loader.add_group_member is False:
            self.abort(
                500,
                "The '{}' backend does not support group member addition"
                "".format(backend)
            )

        ret = []
        status = 200
        for member in members:
            success, message, code = loader.add_group_member(
                name,
                member
            )
            ret.append([code, message])
            status = 201 if success else 200
        return ret, status

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(message="Not allowed to remove member in group")
    @ns.expect(parser)
    @ns.doc(
        responses={
            403: 'Insufficient permissions',
            404: 'No backend found',
        },
    )
    def delete(self, name, backend, member=None):
        """Remove a member from a given group

        **DELETE** method provided by the webservice.
        """
        args = self.parser.parse_args()
        try:
            handler = getattr(bui, 'acl_handler')
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        if backend not in handler.backends:
            self.abort(404, "No acl backend '{}' found".format(backend))

        loader = handler.backends[backend]

        members = [member] if member else (args['memberNames'] or [])

        if loader.del_group_member is False:
            self.abort(
                500,
                "The '{}' backend does not support group member deletion"
                "".format(backend)
            )

        ret = []
        status = 200
        for member in members:
            success, message, code = loader.del_group_member(
                name,
                member
            )
            ret.append([code, message])
            status = 201 if success else 200
        return ret, status


@ns.route('/acl/groupsOf/<member>',
          endpoint='acl_groups_of')
@ns.doc(
    params={
        'member': 'Username',
    }
)
class AclGroupsOf(Resource):
    """The :class:`burpui.api.admin.AclGroupsOf` resource allows you to retrieve
    a list of groups of a given user.

    This resource is part of the :mod:`burpui.api.admin` module.
    """

    @api.acl_admin_or_moderator_required(message="Not allowed to view groups list")
    @ns.marshal_with(groups_list_fields, code=200, description='Success')
    @ns.doc(
        responses={
            403: 'Insufficient permissions',
            404: 'No backend found',
        },
    )
    def get(self, member):
        """Returns a list of group

        **GET** method provided by the webservice.

        :returns: Groups
        """
        return {'groups': [{'name': name, 'inherit': inherit} for name, inherit in meta_grants.get_member_groups(member)]}


@ns.route('/acl/groups',
          '/acl/<backend>/groups',
          '/acl/groups/<name>',
          '/acl/<backend>/groups/<name>',
          endpoint='acl_groups')
@ns.doc(
    params={
        'name': 'Group name',
        'backend': 'Backend name',
    }
)
class AclGroups(Resource):
    """The :class:`burpui.api.admin.AclGroups` resource allows you to
    retrieve a list of groups and to add/update/delete them if your
    acl backend support those actions.

    This resource is part of the :mod:`burpui.api.admin` module.
    """
    parser_add = ns.parser()
    parser_add.add_argument('group', required=True, help='Group name', location='values')
    parser_add.add_argument('grant', required=True, help='Group grant content')
    parser_add.add_argument('backend', help='Backend', location='values')

    parser_mod = ns.parser()
    parser_mod.add_argument('grant', required=True, help='Group grant content')
    parser_mod.add_argument('backend', help='Backend', location='values')

    parser_del = ns.parser()
    parser_del.add_argument('backend', help='Backend', location='values')

    @api.acl_admin_or_moderator_required(message="Not allowed to view groups list")
    @ns.marshal_list_with(group_fields, code=200, description='Success')
    @ns.doc(
        responses={
            403: 'Insufficient permissions',
            404: 'No backend found',
        },
    )
    def get(self, name=None, backend=None):
        """Returns a list of group

        **GET** method provided by the webservice.

        :returns: Groups
        """
        try:
            handler = getattr(bui, 'acl_handler')
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        ret = []
        for _, loader in iteritems(handler.backends):
            groups = loader.groups
            if groups:
                for _id, group in iteritems(groups):
                    append = {
                        'id': _id.lstrip('@'),
                        'grant': group.get('grants', ''),
                        'members': group.get('members', []),
                        'backend': loader.name
                    }
                    if name and name == append['id']:
                        if (backend and backend == append['backend']) or backend is None:
                            return [append]
                    ret.append(append)
        return ret

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(message="Not allowed to create groups")
    @ns.expect(parser_add)
    @ns.doc(
        responses={
            200: 'Request performed with errors',
            201: 'Success',
            403: 'Not allowed',
            400: 'Missing parameters',
            404: 'Backend not found',
            500: 'Backend does not support this operation',
        },
    )
    def put(self, backend=None):
        """Create a new group"""
        args = self.parser_add.parse_args()
        backend = backend or args['backend']

        try:
            handler = getattr(bui, 'acl_handler')
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0 or \
                backend not in handler.backends:
            self.abort(404, "No acl backend found")

        loader = handler.backends[backend]

        if loader.add_group is False:
            self.abort(
                500,
                "The '{}' backend does not support group creation"
                "".format(backend)
            )

        success, message, code = loader.add_group(
            args['group'],
            args['grant']
        )
        status = 201 if success else 200
        return [[code, message]], status

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(message="Not allowed to delete this group")
    @ns.expect(parser_del)
    @ns.doc(
        responses={
            200: 'Request performed with errors',
            201: 'Success',
            403: 'Not allowed',
            400: 'Missing parameters',
            404: 'Backend not found',
            500: 'Backend does not support this operation',
        },
    )
    def delete(self, name, backend=None):
        """Delete a group"""
        args = self.parser_del.parse_args()
        backend = backend or args['backend']

        try:
            handler = getattr(bui, 'acl_handler')
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0 or \
                backend not in handler.backends:
            self.abort(404, "No acl backend found")

        loader = handler.backends[backend]

        if loader.del_group is False:
            self.abort(
                500,
                "The '{}' backend does not support group deletion"
                "".format(backend)
            )

        success, message, code = loader.del_group(
            name
        )
        status = 201 if success else 200
        return [[code, message]], status

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(message="Not allowed to modify this group")
    @ns.expect(parser_mod)
    @ns.doc(
        responses={
            200: 'Request performed with errors',
            201: 'Success',
            403: 'Not allowed',
            400: 'Missing parameters',
            404: 'Backend not found',
            500: 'Backend does not support this operation',
        },
    )
    def post(self, name, backend=None):
        """Change a group"""
        args = self.parser_mod.parse_args()
        backend = backend or args['backend']

        try:
            handler = getattr(bui, 'acl_handler')
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0 or \
                backend not in handler.backends:
            self.abort(404, "No acl backend found")

        loader = handler.backends[backend]

        if loader.mod_group is False:
            self.abort(
                500,
                "The '{}' backend does not support group modification"
                "".format(backend)
            )

        success, message, code = loader.mod_group(
            name,
            args['grant']
        )
        status = 201 if success else 200
        return [[code, message]], status


@ns.route('/acl/grants',
          '/acl/<backend>/grants',
          '/acl/grants/<name>',
          '/acl/<backend>/grants/<name>',
          endpoint='acl_grants')
@ns.doc(
    params={
        'name': 'Grant name',
        'backend': 'Backend name',
    }
)
class AclGrants(Resource):
    """The :class:`burpui.api.admin.AclGrants` resource allows you to
    retrieve a list of grants and to add/update/delete them if your
    acl backend support those actions.

    This resource is part of the :mod:`burpui.api.admin` module.
    """
    parser_add = ns.parser()
    parser_add.add_argument('grant', required=True, help='Grant name')
    parser_add.add_argument('content', required=True, help='Grant content')
    parser_add.add_argument('backend', help='Backend')

    parser_mod = ns.parser()
    parser_mod.add_argument('content', required=True, help='Grant content')
    parser_mod.add_argument('backend', help='Backend')

    parser_del = ns.parser()
    parser_del.add_argument('backend', help='Backend', location='values')

    @api.acl_admin_or_moderator_required(message="Not allowed to view grants list")
    @ns.marshal_list_with(grant_fields, code=200, description='Success')
    @ns.doc(
        responses={
            403: 'Insufficient permissions',
            404: 'No backend found',
        },
    )
    def get(self, name=None, backend=None):
        """Returns a list of grants

        **GET** method provided by the webservice.

        :returns: Grants
        """
        try:
            handler = getattr(bui, 'acl_handler')
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No acl backend found")
        ret = []
        for _, loader in iteritems(handler.backends):
            grants = loader.grants
            if grants:
                for _id, grant in iteritems(grants):
                    append = {
                        'id': _id,
                        'grant': json.dumps(grant),
                        'backend': loader.name
                    }
                    if name and name == _id:
                        if (backend and backend == loader.name) or backend is None:
                            return [append]
                    ret.append(append)
        return ret

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(message="Not allowed to create grants")
    @ns.expect(parser_add)
    @ns.doc(
        responses={
            200: 'Request performed with errors',
            201: 'Success',
            403: 'Not allowed',
            400: 'Missing parameters',
            404: 'Backend not found',
            500: 'Backend does not support this operation',
        },
    )
    def put(self, backend=None):
        """Create a new grant"""
        args = self.parser_add.parse_args()
        backend = backend or args['backend']

        try:
            handler = getattr(bui, 'acl_handler')
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0 or \
                backend not in handler.backends:
            self.abort(404, "No acl backend found")

        loader = handler.backends[backend]

        if loader.add_grant is False:
            self.abort(
                500,
                "The '{}' backend does not support grant creation"
                "".format(backend)
            )

        success, message, code = loader.add_grant(
            args['grant'],
            args['content']
        )
        status = 201 if success else 200
        return [[code, message]], status

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(message="Not allowed to delete this grant")
    @ns.expect(parser_del)
    @ns.doc(
        responses={
            200: 'Request performed with errors',
            201: 'Success',
            403: 'Not allowed',
            400: 'Missing parameters',
            404: 'Backend not found',
            500: 'Backend does not support this operation',
        },
    )
    def delete(self, name, backend=None):
        """Delete a grant"""
        args = self.parser_del.parse_args()
        backend = backend or args['backend']

        try:
            handler = getattr(bui, 'acl_handler')
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0 or \
                backend not in handler.backends:
            self.abort(404, "No acl backend found")

        loader = handler.backends[backend]

        if loader.del_grant is False:
            self.abort(
                500,
                "The '{}' backend does not support grant deletion"
                "".format(backend)
            )

        success, message, code = loader.del_grant(
            name
        )
        status = 201 if success else 200
        return [[code, message]], status

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(message="Not allowed to modify this grant")
    @ns.expect(parser_mod)
    @ns.doc(
        responses={
            200: 'Request performed with errors',
            201: 'Success',
            403: 'Not allowed',
            400: 'Missing parameters',
            404: 'Backend not found',
            500: 'Backend does not support this operation',
        },
    )
    def post(self, name, backend=None):
        """Change a grant"""
        args = self.parser_mod.parse_args()
        backend = backend or args['backend']

        try:
            handler = getattr(bui, 'acl_handler')
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0 or \
                backend not in handler.backends:
            self.abort(404, "No acl backend found")

        loader = handler.backends[backend]

        if loader.mod_grant is False:
            self.abort(
                500,
                "The '{}' backend does not support grant modification"
                "".format(backend)
            )

        success, message, code = loader.mod_grant(
            name,
            args['content']
        )
        status = 201 if success else 200
        return [[code, message]], status


@ns.route('/acl/backend/<backend>', endpoint='acl_backend')
class AclBackend(Resource):
    """The :class:`burpui.api.admin.AclBackend` resource allows you to
    retrieve a given ACL backend with its capabilities.

    This resource is part of the :mod:`burpui.api.admin` module.
    """

    @api.acl_admin_or_moderator_required(message="Not allowed to view backends list")
    @ns.marshal_with(acl_backend_fields, code=200, description='Success')
    @ns.doc(
        responses={
            403: 'Insufficient permissions',
            404: 'No backend found',
        },
    )
    def get(self, backend):
        """Returns a given ACL backend

        **GET** method provided by the webservice.

        :returns: Backend
        """
        try:
            handler = getattr(bui, 'acl_handler')
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No authentication backend found")
        if backend not in handler.backends:
            self.abort(404, "ACL backend {} not found".format(backend))
        loader = handler.backends[backend]
        back = {}
        back['name'] = backend
        back['description'] = gettext(loader.__doc__)
        back['type'] = 'authorization'
        back['priority'] = getattr(loader, 'priority', -1)
        for method in ['add_grant', 'del_grant', 'mod_grant', 'add_group', 'del_group', 'mod_group', 'add_group_member', 'del_group_member', 'add_moderator', 'del_moderator', 'mod_moderator', 'add_admin', 'del_admin']:
            back[method] = getattr(loader, method, False) is not False

        return back


@ns.route('/acl/backends', endpoint='acl_backends')
class AclBackends(Resource):
    """The :class:`burpui.api.admin.AclBackends` resource allows you to
    retrieve a list of ACL backends with their capabilities.

    This resource is part of the :mod:`burpui.api.admin` module.
    """

    @api.acl_admin_or_moderator_required(message="Not allowed to view backends list")
    @ns.marshal_list_with(acl_backend_fields, code=200, description='Success')
    @ns.doc(
        responses={
            403: 'Insufficient permissions',
            404: 'No backend found',
        },
    )
    def get(self):
        """Returns a list of backends

        **GET** method provided by the webservice.

        :returns: Backends
        """
        try:
            handler = getattr(bui, 'acl_handler')
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No authentication backend found")
        ret = []
        for name, backend in iteritems(handler.backends):
            back = {}
            back['name'] = name
            back['description'] = gettext(backend.__doc__)
            back['type'] = 'authorization'
            back['priority'] = getattr(backend, 'priority', -1)
            for method in ['add_grant', 'del_grant', 'mod_grant', 'add_group', 'del_group', 'mod_group', 'add_group_member', 'del_group_member', 'add_moderator', 'del_moderator', 'mod_moderator', 'add_admin', 'del_admin']:
                back[method] = getattr(backend, method, False) is not False
            ret.append(back)

        return ret


@ns.route('/auth/users',
          '/auth/<backend>/users',
          '/auth/users/<name>',
          '/auth/<backend>/users/<name>',
          endpoint='auth_users')
@ns.doc(
    params={
        'name': 'Username',
        'backend': 'Authentication backend',
    }
)
class AuthUsers(Resource):
    """The :class:`burpui.api.admin.AuthUsers` resource allows you to
    retrieve a list of users and to add/update/delete them if your
    authentication backend support those actions.

    This resource is part of the :mod:`burpui.api.admin` module.
    """
    parser_add = ns.parser()
    parser_add.add_argument('username', required=True, help='Username')
    parser_add.add_argument('password', required=True, help='Password')
    parser_add.add_argument('backend', help='Backend')

    parser_mod = ns.parser()
    parser_mod.add_argument('password', required=True, help='Password')
    parser_mod.add_argument('backend', help='Backend')
    parser_mod.add_argument('old_password', required=False, help='Old password')

    parser_del = ns.parser()
    parser_del.add_argument('backend', help='Backend')

    @api.acl_admin_or_moderator_required(message="Not allowed to view users list")
    @ns.marshal_list_with(user_fields, code=200, description='Success')
    @ns.doc(
        responses={
            403: 'Insufficient permissions',
            404: 'No backend found',
        },
    )
    def get(self, name=None, backend=None):
        """Returns a list of users

        **GET** method provided by the webservice.

        :returns: Users
        """
        try:
            handler = getattr(bui, 'uhandler')
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No authentication backend found")
        ret = []
        for _, backend in iteritems(handler.backends):
            loader = backend.loader
            try:
                users = getattr(loader, 'users')
            except AttributeError:
                continue
            if users:
                if isinstance(users, list):
                    for user in users:
                        append = {
                            'id': backend.user(user).get_id(),
                            'name': user,
                            'backend': backend.name
                        }
                        if name and name == append['id']:
                            if (backend and backend == append['backend']) or backend is None:
                                return append
                        ret.append(append)
                elif isinstance(users, dict):
                    for user, _ in iteritems(users):
                        append = {
                            'id': backend.user(user).get_id(),
                            'name': user,
                            'backend': backend.name
                        }
                        if name and name == append['id']:
                            if (backend and backend == append['backend']) or backend is None:
                                return append
                        ret.append(append)
        return ret

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(message="Not allowed to create users")
    @ns.expect(parser_add)
    @ns.doc(
        responses={
            200: 'Request performed with errors',
            201: 'Success',
            403: 'Not allowed',
            400: 'Missing parameters',
            404: 'Backend not found',
            500: 'Backend does not support this operation',
        },
    )
    def put(self, name=None, backend=None):
        """Create a new user"""
        args = self.parser_add.parse_args()
        username = name or args['username']
        backend = backend or args['backend']

        try:
            handler = getattr(bui, 'uhandler')
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0 or \
                backend not in handler.backends:
            self.abort(404, "No authentication backend found")

        backend = handler.backends[backend]

        if backend.add_user is False:
            self.abort(
                500,
                "The '{}' backend does not support user creation"
                "".format(backend)
            )

        success, message, code = backend.add_user(
            username,
            args['password']
        )
        status = 201 if success else 200
        return [[code, message]], status

    @api.disabled_on_demo()
    @api.acl_admin_or_moderator_required(message="Not allowed to delete this user")
    @ns.expect(parser_del)
    @ns.doc(
        responses={
            200: 'Request performed with errors',
            201: 'Success',
            403: 'Not allowed',
            400: 'Missing parameters',
            404: 'Backend not found',
            500: 'Backend does not support this operation',
        },
    )
    def delete(self, name, backend=None):
        """Delete a user"""
        args = self.parser_del.parse_args()
        backend = backend or args['backend']

        try:
            handler = getattr(bui, 'uhandler')
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0 or \
                backend not in handler.backends:
            self.abort(404, "No authentication backend found")

        backend = handler.backends[backend]

        if backend.del_user is False:
            self.abort(
                500,
                "The '{}' backend does not support user deletion"
                "".format(backend)
            )

        success, message, code = backend.del_user(
            name
        )
        status = 201 if success else 200
        return [[code, message]], status

    @api.disabled_on_demo()
    @api.acl_own_or_admin(key='name', message="Not allowed to modify this user")
    @ns.expect(parser_mod)
    @ns.doc(
        responses={
            200: 'Request performed with errors',
            201: 'Success',
            403: 'Not allowed',
            400: 'Missing parameters',
            404: 'Backend not found',
            500: 'Backend does not support this operation',
        },
    )
    def post(self, name, backend=None):
        """Change user password"""
        args = self.parser_mod.parse_args()
        backend = backend or args['backend']
        is_admin = True

        if not current_user.is_anonymous:
            is_admin = current_user.acl.is_admin()

        if not is_admin and not args['old_password']:
            self.abort(400, "Old password required")

        try:
            handler = getattr(bui, 'uhandler')
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0 or \
                backend not in handler.backends:
            self.abort(404, "No authentication backend found")

        backend = handler.backends[backend]

        if backend.change_password is False:
            self.abort(
                500,
                "The '{}' backend does not support user modification"
                "".format(backend)
            )

        success, message, code = backend.change_password(
            name,
            args['password'],
            args.get('old_password')
        )
        status = 201 if success else 200
        return [[code, message]], status


@ns.route('/auth/backend/<backend>', endpoint='auth_backend')
class AuthBackend(Resource):
    """The :class:`burpui.api.admin.AuthBackend` resource allows you to
    retrieve a given authentication backend and its capabilities.

    This resource is part of the :mod:`burpui.api.admin` module.
    """

    @api.acl_admin_or_moderator_required(message="Not allowed to view backends list")
    @ns.marshal_with(auth_backend_fields, code=200, description='Success')
    @ns.doc(
        responses={
            403: 'Insufficient permissions',
            404: 'No backend found',
        },
    )
    def get(self, backend):
        """Returns a given authentication backend

        **GET** method provided by the webservice.

        :returns: Backend
        """
        try:
            handler = getattr(bui, 'uhandler')
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No authentication backend found")
        if backend not in handler.backends:
            self.abort(404, "No authentication backend {} found".format(backend))
        back = handler.backends[backend]
        ret = {
            'name': backend,
            'description': gettext(back.__doc__),
            'type': 'authentication',
            'priority': getattr(back, 'priority', -1),
            'add': getattr(back, 'add_user', False) is not False,
            'del': getattr(back, 'del_user', False) is not False,
            'mod': getattr(back, 'change_password', False) is not False,
        }

        return ret


@ns.route('/auth/backends', endpoint='auth_backends')
class AuthBackends(Resource):
    """The :class:`burpui.api.admin.AuthBackends` resource allows you to
    retrieve a list of backends and to add/update/delete users if your
    authentication backend support those actions.

    This resource is part of the :mod:`burpui.api.admin` module.
    """

    @api.acl_admin_or_moderator_required(message="Not allowed to view backends list")
    @ns.marshal_list_with(auth_backend_fields, code=200, description='Success')
    @ns.doc(
        responses={
            403: 'Insufficient permissions',
            404: 'No backend found',
        },
    )
    def get(self):
        """Returns a list of backends

        **GET** method provided by the webservice.

        :returns: Backends
        """
        try:
            handler = getattr(bui, 'uhandler')
        except AttributeError:
            handler = None

        if not handler or len(handler.backends) == 0:
            self.abort(404, "No authentication backend found")
        ret = []
        for name, backend in iteritems(handler.backends):
            ret.append({
                'name': name,
                'description': gettext(backend.__doc__),
                'type': 'authentication',
                'priority': backend.priority,
                'add': backend.add_user is not False,
                'del': backend.del_user is not False,
                'mod': backend.change_password is not False,
            })

        return ret


@ns.route('/session/<user>',
          '/session/<user>/<uuid:id>',
          endpoint='other_sessions')
@ns.doc(
    params={
        'user': 'User to get sessions from',
        'id': 'Session id',
    }
)
class OtherSessions(Resource):
    """The :class:`burpui.api.admin.OtherSessions` resource allows you to
    retrieve a list of sessions for a given user.

    This resource is part of the :mod:`burpui.api.admin` module.
    """

    @ns.marshal_list_with(session_fields, code=200, description='Success')
    @ns.doc(
        responses={
            403: 'Insufficient permissions',
            404: 'User not found',
        },
    )
    def get(self, user=None, id=None):
        """Returns a list of sessions

        **GET** method provided by the webservice.

        :returns: Sessions
        """
        if id:
            return session_manager.get_session_by_id(str(id))
        if not user:
            self.abort(404, 'User not found')
        return session_manager.get_user_sessions(user)


@ns.route('/me/session',
          '/me/session/<uuid:id>',
          endpoint='user_sessions')
@ns.doc(
    params={
        'id': 'Session id',
    }
)
class MySessions(Resource):
    """The :class:`burpui.api.admin.MySessions` resource allows you to
    retrieve a list of sessions and invalidate them for the current user.

    This resource is part of the :mod:`burpui.api.admin` module.
    """

    @ns.marshal_list_with(session_fields, code=200, description='Success')
    @ns.doc(
        responses={
            403: 'Insufficient permissions',
            404: 'User not found',
        },
    )
    def get(self, id=None):
        """Returns a list of sessions

        **GET** method provided by the webservice.

        :returns: Sessions
        """
        if id:
            return session_manager.get_session_by_id(str(id))
        user = getattr(current_user, 'name', None)
        if not user:
            self.abort(404, 'User not found')
        return session_manager.get_user_sessions(user)

    @api.disabled_on_demo()
    @ns.doc(
        responses={
            201: 'Success',
            403: 'Insufficient permissions',
            404: 'User or session not found',
            400: 'Wrong request'
        }
    )
    def delete(self, id=None):
        """Delete a given session

        Note: ``id`` is mandatory
        """
        if not id:
            self.abort(400, 'Missing id')
        user = getattr(current_user, 'name', None)
        if not user:
            self.abort(404, 'User not found')
        store = session_manager.get_session_by_id(str(id))
        if not store:
            self.abort('Session not found')
        if store.user != user:
            if not current_user.is_anonymous and \
                    not current_user.acl.is_admin() and \
                    not current_user.acl.is_moderator():
                self.abort(403, 'Insufficient permissions')
            if current_user.acl.is_moderator() and \
                    meta_grants.is_admin(store.user):
                self.abort(403, 'Insufficient permissions')
        if session_manager.invalidate_session_by_id(store.uuid):
            session_manager.delete_session_by_id(store.uuid)
        return [NOTIF_OK, 'Session {} successfully revoked'.format(id)], 201



# -*- coding: utf8 -*-
"""
.. module:: burpui.api.async
    :platform: Unix
    :synopsis: Burp-UI async api module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
import os
import select
import struct

from . import api, cache_key, force_refresh
from .misc import History
from .custom import Resource
from .client import node_fields
from .clients import RunningBackup, ClientsReport, RunningClients
from ..server import BUIServer  # noqa
from ..ext.cache import cache
from ..config import config
from ..decorators import browser_cache
from ..tasks import perform_restore, load_all_tree

from time import time
from zlib import adler32
from flask import url_for, Response, current_app, after_this_request, \
    send_file, request
from flask_login import current_user
from datetime import timedelta
from werkzeug.datastructures import Headers
try:
    from .ext.ws import socketio  # noqa
    WS_AVAILABLE = True
except ImportError:
    WS_AVAILABLE = False

if config.get('WITH_SQL'):
    from ..ext.sql import db
    from ..models import Task
else:
    db = None

bui = current_app  # type: BUIServer
ns = api.namespace('async', 'Asynchronous methods')

task_types = {
    'restore': (perform_restore, '.async_get_file'),
    'browse': (load_all_tree, '.async_do_browse_all'),
}


@ns.route('/status/<task_type>/<task_id>',
          '/<server>/status/<task_type>/<task_id>',
          endpoint='async_status')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
        'task_id': 'The task ID to process',
        'task_type': 'The task type (either "restore" or "browse")',
    }
)
class AsyncStatus(Resource):
    """The :class:`burpui.api.async.AsyncStatus` resource allows you to
    follow a restore task.

    This resource is part of the :mod:`burpui.api.async` module.
    """
    if config['WITH_LIMIT']:
        try:
            from ..ext.limit import limiter
            decorators = [limiter.exempt]
        except ImportError:
            pass

    @ns.doc(
        responses={
            200: 'Success',
            500: 'Task failed',
        },
    )
    def get(self, task_type, task_id, server=None):
        """Returns the state of the given task"""
        if task_type not in task_types:
            return {'state': 'FAILURE'}
        task_obj, callback = task_types[task_type]
        task = task_obj.AsyncResult(task_id)
        if task.state == 'FAILURE':
            if db:
                rec = Task.query.filter_by(uuid=task_id).first()
                if rec:
                    try:
                        db.session.delete(rec)
                        db.session.commit()
                    except:
                        db.session.rollback()
            task.revoke()
            err = str(task.result)
            self.abort(502, err)
        if task.state == 'SUCCESS':
            if not task.result:
                self.abort(500, 'The task did not return anything')
            server = task.result.get('server')
            return {
                'state': task.state,
                'location': url_for(
                    callback,
                    task_id=task_id,
                    server=server
                )
            }
        return {'state': task.state}

    @ns.doc(
        responses={
            201: 'Success',
            400: 'Wrong request',
            403: 'Permission denied',
        },
    )
    def delete(self, task_type, task_id, server=None):
        """Cancel a given task"""
        if task_type not in task_types:
            return '', 400
        task_obj, _ = task_types[task_type]
        task = task_obj.AsyncResult(task_id)
        user = task.result.get('user')
        dst_server = task.result.get('server')

        if (current_user.name != user or (dst_server and dst_server != server)) and \
                not current_user.acl.is_admin():
            self.abort(403, 'Unauthorized access')

        # do not remove the task from db yet since we may need to remove
        # some temporary files afterward. The "cleanup_restore" task will take
        # care of this
        task.revoke()
        return '', 201


@ns.route('/get/<task_id>',
          '/<server>/get/<task_id>',
          endpoint='async_get_file')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
        'task_id': 'The task ID to process',
    }
)
class AsyncGetFile(Resource):
    """The :class:`burpui.api.async.AsyncGetFile` resource allows you to
    retrieve the archive generated by the given task.

    This resource is part of the :mod:`burpui.api.async` module.
    """

    @ns.doc(
        responses={
            200: 'Success',
            400: 'Incomplete task',
            403: 'Insufficient permissions',
            500: 'Task failed',
        },
    )
    def get(self, task_id, server=None):
        """Returns the generated archive"""
        task = perform_restore.AsyncResult(task_id)
        if task.state != 'SUCCESS':
            if task.state == 'FAILURE':
                err = task.result.get('error')
                if err != 'encrypted' and not task.result.get('admin'):
                    err = 'An error occurred while performing the ' \
                          'restoration. Please contact your administrator ' \
                          'for further details'
                self.abort(
                    500,
                    'Unsuccessful task:\n{}'.format(err)
                )
            self.abort(400, 'Task not processed yet: {}'.format(task.state))

        path = task.result.get('path')
        user = task.result.get('user')
        dst_server = task.result.get('server')
        filename = task.result.get('filename')

        if (current_user.name != user or (dst_server and dst_server != server)) and \
                not current_user.acl.is_admin():
            self.abort(403, 'Unauthorized access')

        if db:
            rec = Task.query.filter_by(uuid=task_id).first()
            if rec:
                try:
                    db.session.delete(rec)
                    db.session.commit()
                except:
                    db.session.rollback()
        task.revoke()

        if dst_server:
            return self.stream_file(path, filename, dst_server)

        try:
            # Trick to delete the file while sending it to the client.
            # First, we open the file in reading mode so that a file handler
            # is open on the file. Then we delete it as soon as the request
            # ended. Because the fh is open, the file will be actually removed
            # when the transfer is done and the send_file method has closed
            # the fh. Only tested on Linux systems.
            fh = open(path, 'rb')

            @after_this_request
            def remove_file(response):
                """Callback function to run after the client has handled
                the request to remove temporary files.
                """
                os.remove(path)
                return response

            resp = send_file(fh,
                             as_attachment=True,
                             attachment_filename=filename,
                             mimetype='application/zip')
            resp.set_cookie('fileDownload', 'true')
        except Exception as e:
            bui.client.logger.error(str(e))
            self.abort(500, str(e))

        return resp

    def stream_file(self, path, filename, server):
        socket = bui.client.get_file(path, server)
        if not socket:
            self.abort(500)
        lengthbuf = socket.recv(8)
        length, = struct.unpack('!Q', lengthbuf)

        def stream(sock, l):
            """The restoration took place on another server so we need
            to stream the file that is not present on the current
            machine.
            """
            bsize = 1024
            received = 0
            if l < bsize:
                bsize = l
            while received < l:
                buf = b''
                r, _, _ = select.select([sock], [], [], 5)
                if not r:
                    self.abort(500, 'Socket timed-out')
                buf += sock.recv(bsize)
                if not buf:
                    continue
                received += len(buf)
                self.logger.debug('{}/{}'.format(received, l))
                yield buf
            sock.sendall(struct.pack('!Q', 2))
            sock.sendall(b'RE')
            sock.close()

        headers = Headers()
        headers.add('Content-Disposition',
                    'attachement',
                    filename=filename)
        headers['Content-Length'] = length

        resp = Response(stream(socket, length),
                        mimetype='application/zip',
                        headers=headers,
                        direct_passthrough=True)
        resp.set_cookie('fileDownload', 'true')
        resp.set_etag('flask-%s-%s-%s' % (
            time(),
            length,
            adler32(filename.encode('utf-8')) & 0xffffffff))

        return resp


@ns.route('/archive/<name>/<int:backup>',
          '/<server>/archive/<name>/<int:backup>',
          endpoint='async_restore')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
        'name': 'Client name',
        'backup': 'Backup number',
    },
)
class AsyncRestore(Resource):
    """The :class:`burpui.api.async.AsyncRestore` resource allows you to
    perform a file restoration.

    This resource is part of the :mod:`burpui.api.async` module.

    The following parameters are supported:
    - ``list``: list of files/directories to restore
    - ``strip``: number of elements to strip in the path
    - ``format``: returning archive format
    - ``pass``: password to use for encrypted backups
    """
    parser = ns.parser()
    parser.add_argument(
        'pass',
        help='Password to use for encrypted backups',
        nullable=True
    )
    parser.add_argument(
        'format',
        required=False,
        help='Returning archive format',
        choices=('zip', 'tar.gz', 'tar.bz2'),
        default='zip',
        nullable=True
    )
    parser.add_argument(
        'strip',
        type=int,
        help='Number of elements to strip in the path',
        default=0,
        nullable=True
    )
    parser.add_argument(
        'list',
        required=True,
        help='List of files/directories to restore',
        nullable=False
    )

    @ns.expect(parser, validate=True)
    @ns.doc(
        responses={
            202: 'Accepted',
            400: 'Missing parameter',
            403: 'Insufficient permissions',
        },
    )
    def post(self, server=None, name=None, backup=None):
        """Performs an online restoration

        **POST** method provided by the webservice.

        This method returns a :mod:`flask.Response` object.

        :param server: Which server to collect data from when in multi-agent
                       mode
        :type server: str

        :param name: The client we are working on
        :type name: str

        :param backup: The backup we are working on
        :type backup: int
        """
        args = self.parser.parse_args()
        files = args['list']
        strip = args['strip']
        fmt = args['format'] or 'zip'
        passwd = args['pass']
        room = None
        if WS_AVAILABLE:
            room = request.sid
        if not files or not name or not backup:
            self.abort(400, 'missing arguments')
        # Manage ACL
        if not current_user.is_anonymous and \
                not current_user.acl.is_admin() and \
                not current_user.acl.is_client_rw(name, server):
            self.abort(
                403,
                'You are not allowed to perform a restoration for this client'
            )
        task = perform_restore.apply_async(
            args=[
                name,
                backup,
                files,
                strip,
                fmt,
                passwd,
                server,
                current_user.name,
                not current_user.is_anonymous and current_user.acl.is_admin(),
                room
            ]
        )
        if db:
            db_task = Task(
                task.id,
                'perform_restore',
                current_user.name,
                timedelta(minutes=60)
            )
            try:
                db.session.add(db_task)
                db.session.commit()
            except:
                db.session.rollback()
        return {'id': task.id, 'name': 'perform_restore'}, 202


@ns.route('/running',
          '/<server>/running',
          '/running/<client>',
          '/<server>/running/<client>',
          endpoint='async_running_clients')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
        'client': 'Client name',
    },
)
class AsyncRunningClients(RunningClients):
    """The :class:`burpui.api.async.AsyncRunningClients` resource allows you
    to retrieve a list of clients that are currently running a backup.

    This resource is part of the :mod:`burpui.api.async` module.

    This resource is backed by a periodic task. If the periodic task fail or is
    not running, we fallback to the "synchronous" API call.

    An optional ``GET`` parameter called ``serverName`` is supported when running
    in multi-agent mode.

    .. seealso:: :class:`burpui.api.clients.RunningClients`
    """

    def get(self, client=None, server=None):
        """Returns a list of clients currently running a backup

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            [ 'client1', 'client2' ]


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to see.

        :param server: Which server to collect data from when in multi-agent mode
        :type server: str

        :param client: Ask a specific client in order to know if it is running a backup
        :type client: str

        :returns: The *JSON* described above.
        """
        server = server or self.parser.parse_args()['serverName']
        res = cache.cache.get('backup_running_result')
        if res is None:
            res = bui.client.is_one_backup_running(server)
        return self._running_clients(res, client, server)


@ns.route('/backup-running',
          '/<server>/backup-running',
          endpoint='async_running_backup')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
    }
)
class AsyncRunningBackup(RunningBackup):
    """The :class:`burpui.api.async.AsyncRunningBackup` resource allows you to
    access the status of the server in order to know if there is a running
    backup currently.

    This resource is backed by a periodic task. If the periodic task fail or is
    not running, we fallback to the "synchronous" API call.

    This resource is part of the :mod:`burpui.api.async` module.
    """

    @ns.marshal_with(
        RunningBackup.running_fields,
        code=200,
        description='Success',
        strict=False
    )
    def get(self, server=None):
        """Tells if a backup is running right now

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            {
                "running": false
            }


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent
                       mode
        :type server: str

        :returns: The *JSON* described above.
        """
        res = cache.cache.get('backup_running_result')
        if res is None:
            res = bui.client.is_one_backup_running(server)
        return {'running': self._is_one_backup_running(res, server)}


@ns.route('/history',
          '/history/<client>',
          '/<server>/history',
          '/<server>/history/<client>',
          endpoint='async_history')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
        'client': 'Client name',
    },
)
class AsyncHistory(History):
    """The :class:`burpui.api.misc.History` resource allows you to retrieve
    an history of the backups

    This resource is backed by a periodic task. If the periodic task fail or is
    not running, we redirect to the "synchronous" API call.

    An optional ``GET`` parameter called ``serverName`` is supported when
    running in multi-agent mode and ``clientName`` is also allowed to filter
    by client.

    ::

        $('#calendar').fullCalendar({

            eventSources: [

                // your event source
                {
                    events: [ // put the array in the `events` property
                        {
                            title  : 'event1',
                            start  : '2010-01-01'
                        },
                        {
                            title  : 'event2',
                            start  : '2010-01-05',
                            end    : '2010-01-07'
                        },
                        {
                            title  : 'event3',
                            start  : '2010-01-09T12:30:00',
                        }
                    ],
                    color: 'black',     // an option!
                    textColor: 'yellow' // an option!
                }

                // any other event sources...

            ]

        });

    """

    @cache.cached(timeout=1800, key_prefix=cache_key, unless=force_refresh)
    @ns.marshal_with(
        History.history_fields,
        code=200,
        description='Success',
        strict=False,
        as_list=True
    )
    @ns.expect(History.parser)
    @ns.doc(
        responses={
            200: 'Success',
            403: 'Insufficient permissions',
        },
    )
    @browser_cache(1800)
    def get(self, client=None, server=None):
        """Returns a list of calendars describing the backups that have been
        completed so far

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            [
              {
                "color": "#7C6F44",
                "events": [
                  {
                    "backup": "0000001",
                    "end": "2015-01-25 13:32:04+01:00",
                    "name": "toto-test",
                    "start": "2015-01-25 13:32:00+01:00",
                    "title": "Client: toto-test, Backup n°0000001",
                    "url": "/client/toto-test"
                  }
                ],
                "name": "toto-test",
                "textColor": "white"
              }
            ]


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent mode
        :type server: str
        :param client: Which client to collect data from
        :type client: str

        :returns: The *JSON* described above
        """
        self._check_acl(client, server)
        res = cache.cache.get('all_backups')
        if res is None:
            return self._get_backup_history(client, server)

        return self._get_backup_history(client, server, res)


@ns.route('/report',
          '/<server>/report',
          endpoint='async_clients_report')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
    },
)
class AsyncClientsReport(ClientsReport):
    """The :class:`burpui.api.async.AsyncClientsReport` resource allows you to
    access general reports about your clients.

    This resource is part of the :mod:`burpui.api.clients` module.

    An optional ``GET`` parameter called ``serverName`` is supported when
    running in multi-agent mode.
    """

    @cache.cached(timeout=1800, key_prefix=cache_key, unless=force_refresh)
    @ns.marshal_with(
        ClientsReport.report_fields,
        code=200,
        description='Success',
        strict=False
    )
    @ns.expect(ClientsReport.parser)
    @ns.doc(
        responses={
            403: 'Insufficient permissions',
            500: 'Internal failure',
        },
    )
    @browser_cache(1800)
    def get(self, server=None):
        """Returns a global report about all the clients of a given server

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            {
              "backups": [
                {
                  "name": "client1",
                  "number": 15
                },
                {
                  "name": "client2",
                  "number": 1
                }
              ],
              "clients": [
                {
                  "name": "client1",
                  "stats": {
                    "total": 296377,
                    "totsize": 57055793698,
                    "windows": "unknown"
                  }
                },
                {
                  "name": "client2",
                  "stats": {
                    "total": 3117,
                    "totsize": 5345361,
                    "windows": "true"
                  }
                }
              ]
            }


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent
                       mode
        :type server: str

        :returns: The *JSON* described above
        """
        server = server or self.parser.parse_args()['serverName']
        self._check_acl(server)
        res = cache.cache.get('all_clients_reports')
        if res is None:
            return self._get_clients_reports(server=server)

        return self._get_clients_reports(res, server)


@ns.route('/browseall/<name>/<int:backup>',
          '/<server>/browsall/<name>/<int:backup>',
          endpoint='async_client_tree_all')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in' +
                  ' multi-agent mode',
        'name': 'Client name',
        'backup': 'Backup number',
    },
)
class AsyncClientTreeAll(Resource):
    """The :class:`burpui.api.async.AsyncClientTreeAll` resource allows you to
    retrieve a list of all the files in a given backup through the celery
    worker.

    This resource is part of the :mod:`burpui.api.client` module.

    An optional ``GET`` parameter called ``serverName`` is supported when
    running in multi-agent mode.
    """
    parser = ns.parser()
    parser.add_argument(
        'serverName',
        help='Which server to collect data from when in multi-agent mode'
    )

    @ns.expect(parser)
    @ns.doc(
        responses={
            202: 'Accepted',
            405: 'Method not allowed',
            403: 'Insufficient permissions',
            500: 'Internal failure',
        },
    )
    def post(self, server=None, name=None, backup=None):
        """Launch the tasks that will gather all nodes of a given backup

        **POST** method provided by the webservice.

        This method returns a :mod:`flask.Response` object.

        :param server: Which server to collect data from when in multi-agent
                       mode
        :type server: str

        :param name: The client we are working on
        :type name: str

        :param backup: The backup we are working on
        :type backup: int
        """
        args = self.parser.parse_args()
        server = server or args.get('serverName')

        if not bui.client.get_attr('batch_list_supported', False, server):
            self.abort(
                405,
                'Sorry, the requested backend does not support this method'
            )

        # Manage ACL
        if not current_user.is_anonymous and \
                not current_user.acl.is_admin() and \
                not current_user.acl.is_client_allowed(name, server):
            self.abort(403, 'Sorry, you are not allowed to view this client')

        task = load_all_tree.apply_async(
            args=[
                name,
                backup,
                server,
                current_user.name
            ]
        )
        return {'id': task.id, 'name': 'load_all_tree'}, 202


@ns.route('/get-browse/<task_id>',
          '/<server>/get-browse/<task_id>',
          endpoint='async_do_browse_all')
@ns.doc(
    params={
        'task_id': 'The task ID to process',
    }
)
class AsyncDoBrowseAll(Resource):
    """The :class:`burpui.api.async.AsyncDoBrowseAll` resource allows you to
    retrieve the tree generated by the given task.

    This resource is part of the :mod:`burpui.api.async` module.
    """

    @ns.marshal_list_with(node_fields, code=200, description='Success')
    @ns.doc(
        responses={
            400: 'Incomplete task',
            403: 'Insufficient permissions',
            500: 'Task failed',
        },
    )
    def get(self, task_id, server=None):
        """Returns the generated archive"""
        task = load_all_tree.AsyncResult(task_id)
        if task.state != 'SUCCESS':
            if task.state == 'FAILURE':
                self.abort(
                    500,
                    'Unsuccessful task: {}'.format(task.result.get('error'))
                )
            self.abort(400, 'Task not processed yet: {}'.format(task.state))

        user = task.result.get('user')
        dst_server = task.result.get('server')
        resp = task.result.get('tree')

        if current_user.name != user or (dst_server and dst_server != server):
            self.abort(403, 'Unauthorized access')

        task.revoke()

        return resp
